#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import TYPE_CHECKING, Optional

from neptune.new.attributes.constants import MONITORING_STDOUT_ATTRIBUTE_PATH, MONITORING_STDERR_ATTRIBUTE_PATH
from neptune.new.internal.background_job import BackgroundJob
from neptune.new.internal.streams.std_stream_capture_logger import StdoutCaptureLogger, StderrCaptureLogger

if TYPE_CHECKING:
    from neptune.new.run import Run


class StdoutCaptureBackgroundJob(BackgroundJob):

    def __init__(self, attribute_name: str = MONITORING_STDOUT_ATTRIBUTE_PATH):
        self._attribute_name = attribute_name
        self._logger = None

    def start(self, run: 'Run'):
        self._logger = StdoutCaptureLogger(run, self._attribute_name)

    def stop(self):
        self._logger.close()

    def join(self, seconds: Optional[float] = None):
        pass


class StderrCaptureBackgroundJob(BackgroundJob):

    def __init__(self, attribute_name: str = MONITORING_STDERR_ATTRIBUTE_PATH):
        self._attribute_name = attribute_name
        self._logger = None

    def start(self, run: 'Run'):
        self._logger = StderrCaptureLogger(run, self._attribute_name)

    def stop(self):
        self._logger.close()

    def join(self, seconds: Optional[float] = None):
        pass



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import threading

import six


class NeptuneThread(threading.Thread):
    def __init__(self, is_daemon):
        super(NeptuneThread, self).__init__(target=self.run)
        self.setDaemon(is_daemon)
        self._interrupted = threading.Event()

    def should_continue_running(self):
        # TODO: remove this pylint exception once we stop supporting Python 2
        # pylint: disable=no-member
        if six.PY2:
            all_threads = threading.enumerate()

            # pylint:disable=protected-access
            main_thread_is_alive = any(t.__class__ is threading._MainThread and t.is_alive() for t in all_threads)
        else:
            main_thread_is_alive = threading.main_thread().is_alive()

        return not self._interrupted.is_set() and main_thread_is_alive

    def interrupt(self):
        self._interrupted.set()

    def run(self):
        raise NotImplementedError()



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Union, Optional, Iterable

import click

from neptune.new.attributes.series.fetchable_series import FetchableSeries
from neptune.new.internal.backends.api_model import FloatSeriesValues
from neptune.new.types.series.float_series import FloatSeries as FloatSeriesVal

from neptune.new.internal.utils import verify_type

from neptune.new.internal.operation import ClearFloatLog, LogFloats, Operation, ConfigFloatSeries
from neptune.new.attributes.series.series import Series

Val = FloatSeriesVal
Data = Union[float, int]


class FloatSeries(Series[Val, Data], FetchableSeries[FloatSeriesValues]):

    # pylint: disable=redefined-builtin
    def configure(self,
                  min: Optional[Union[float, int]] = None,
                  max: Optional[Union[float, int]] = None,
                  unit: Optional[str] = None,
                  wait: bool = False) -> None:
        verify_type("min", min, (float, int))
        verify_type("max", max, (float, int))
        verify_type("unit", unit, str)
        with self._run.lock():
            self._enqueue_operation(ConfigFloatSeries(self._path, min, max, unit), wait)

    def _get_log_operation_from_value(self, value: Val, step: Optional[float], timestamp: float) -> Operation:
        values = [LogFloats.ValueType(val, step=step, ts=timestamp) for val in value.values]
        return LogFloats(self._path, values)

    def _get_clear_operation(self) -> Operation:
        return ClearFloatLog(self._path)

    def _get_config_operation_from_value(self, value: Val) -> Optional[Operation]:
        return ConfigFloatSeries(self._path, value.min, value.max, value.unit)

    def _data_to_value(self, values: Iterable, **kwargs) -> Val:
        if kwargs:
            click.echo("Warning: unexpected arguments ({kwargs}) in FloatSeries".format(kwargs=kwargs), err=True)
        return FloatSeriesVal(values)

    def _is_value_type(self, value) -> bool:
        return isinstance(value, FloatSeriesVal)

    def fetch_last(self) -> float:
        # pylint: disable=protected-access
        val = self._backend.get_float_series_attribute(self._run_uuid, self._path)
        return val.last

    def _fetch_values_from_backend(self, offset, limit) -> FloatSeriesValues:
        return self._backend.get_float_series_values(self._run_uuid, self._path, offset, limit)



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import uuid
from datetime import datetime
from shutil import copyfile
from typing import Optional, List, Dict, TypeVar, Type, Tuple, Any
from zipfile import ZipFile

from neptune.new.exceptions import (
    RunNotFound,
    RunUUIDNotFound,
    InternalClientError,
    MetadataInconsistency,
    NeptuneException,
)
from neptune.new.internal.backends.api_model import (
    ApiRun,
    Attribute,
    AttributeType,
    BoolAttribute,
    DatetimeAttribute,
    FileAttribute,
    FloatAttribute,
    FloatSeriesAttribute,
    IntAttribute,
    Project,
    Workspace,
    StringAttribute,
    StringSeriesAttribute,
    StringSetAttribute,
    StringSeriesValues,
    FloatSeriesValues,
    ImageSeriesValues,
    StringPointValue,
)
from neptune.new.internal.backends.hosted_file_operations import get_unique_upload_entries
from neptune.new.internal.backends.neptune_backend import NeptuneBackend
from neptune.new.internal.run_structure import RunStructure
from neptune.new.internal.operation import (
    AddStrings,
    AssignBool, AssignDatetime,
    AssignFloat,
    AssignInt, AssignString,
    ClearFloatLog,
    ClearImageLog,
    ClearStringLog,
    ClearStringSet,
    ConfigFloatSeries,
    DeleteAttribute,
    LogFloats,
    LogImages,
    LogStrings,
    Operation,
    RemoveStrings,
    UploadFile,
    UploadFileContent,
    UploadFileSet, DeleteFiles,
)
from neptune.new.internal.operation_visitor import OperationVisitor
from neptune.new.internal.utils import base64_decode
from neptune.new.internal.utils.generic_attribute_mapper import NoValue
from neptune.new.internal.utils.paths import path_to_str
from neptune.new.types import Boolean, Integer
from neptune.new.types.atoms import GitRef
from neptune.new.types.atoms.datetime import Datetime
from neptune.new.types.atoms.file import File
from neptune.new.types.atoms.float import Float
from neptune.new.types.atoms.string import String
from neptune.new.types.file_set import FileSet
from neptune.new.types.namespace import Namespace
from neptune.new.types.series.float_series import FloatSeries
from neptune.new.types.series.file_series import FileSeries
from neptune.new.types.series.string_series import StringSeries
from neptune.new.types.sets.string_set import StringSet
from neptune.new.types.value import Value
from neptune.new.types.value_visitor import ValueVisitor

Val = TypeVar('Val', bound=Value)


class NeptuneBackendMock(NeptuneBackend):

    def __init__(self, credentials=None, proxies=None):
        # pylint: disable=unused-argument
        self._runs: Dict[uuid.UUID, RunStructure[Value, dict]] = dict()
        self._attribute_type_converter_value_visitor = self.AttributeTypeConverterValueVisitor()

    def get_display_address(self) -> str:
        return "OFFLINE"

    def get_project(self, project_id: str) -> Project:
        return Project(uuid.uuid4(), "project-placeholder", "offline")

    def get_available_projects(self,
                               workspace_id: Optional[str] = None,
                               search_term: Optional[str] = None
                               ) -> List[Project]:
        return [Project(uuid.uuid4(), "project-placeholder", "offline")]

    def get_available_workspaces(self) -> List[Workspace]:
        return [Workspace(uuid.uuid4(), "offline")]

    def create_run(self,
                   project_uuid: uuid.UUID,
                   git_ref: Optional[GitRef] = None,
                   custom_run_id: Optional[str] = None,
                   notebook_id: Optional[uuid.UUID] = None,
                   checkpoint_id: Optional[uuid.UUID] = None
                   ) -> ApiRun:
        short_id = "OFFLINE-{}".format(len(self._runs) + 1)
        new_run_uuid = uuid.uuid4()
        self._runs[new_run_uuid] = RunStructure[Value, dict]()
        self._runs[new_run_uuid].set(["sys", "id"], String(short_id))
        self._runs[new_run_uuid].set(["sys", "state"], String("running"))
        self._runs[new_run_uuid].set(["sys", "owner"], String("offline_user"))
        self._runs[new_run_uuid].set(["sys", "size"], Float(0))
        self._runs[new_run_uuid].set(["sys", "tags"], StringSet(set()))
        self._runs[new_run_uuid].set(["sys", "creation_time"], Datetime(datetime.now()))
        self._runs[new_run_uuid].set(["sys", "modification_time"], Datetime(datetime.now()))
        self._runs[new_run_uuid].set(["sys", "failed"], Boolean(False))
        if git_ref:
            self._runs[new_run_uuid].set(["source_code", "git"], git_ref)
        return ApiRun(new_run_uuid, short_id, 'workspace', 'sandbox', False)

    def create_checkpoint(self, notebook_id: uuid.UUID, jupyter_path: str) -> Optional[uuid.UUID]:
        return None

    def get_run(self, run_id: str) -> ApiRun:
        raise RunNotFound(run_id)

    def execute_operations(self, run_uuid: uuid.UUID, operations: List[Operation]) -> List[NeptuneException]:
        result = []
        for op in operations:
            try:
                self._execute_operation(run_uuid, op)
            except NeptuneException as e:
                result.append(e)
        return result

    def _execute_operation(self, run_uuid: uuid.UUID, op: Operation) -> None:
        if run_uuid not in self._runs:
            raise RunUUIDNotFound(run_uuid)
        run = self._runs[run_uuid]
        val = run.get(op.path)
        if val is not None and not isinstance(val, Value):
            if isinstance(val, dict):
                raise MetadataInconsistency("{} is a namespace, not an attribute".format(op.path))
            else:
                raise InternalClientError("{} is a {}".format(op.path, type(val)))
        visitor = NeptuneBackendMock.NewValueOpVisitor(op.path, val)
        new_val = visitor.visit(op)
        if new_val is not None:
            run.set(op.path, new_val)
        else:
            run.pop(op.path)

    def get_attributes(self, run_uuid: uuid.UUID) -> List[Attribute]:
        if run_uuid not in self._runs:
            raise RunUUIDNotFound(run_uuid)
        run = self._runs[run_uuid]
        return list(self._generate_attributes(None, run.get_structure()))

    def _generate_attributes(self, base_path: Optional[str], values: dict):
        for key, value_or_dict in values.items():
            new_path = base_path + '/' + key if base_path is not None else key
            if isinstance(value_or_dict, dict):
                yield from self._generate_attributes(new_path, value_or_dict)
            else:
                yield Attribute(new_path, value_or_dict.accept(self._attribute_type_converter_value_visitor))

    def download_file(self, run_uuid: uuid.UUID, path: List[str], destination: Optional[str] = None):
        value: File = self._runs[run_uuid].get(path)
        target_path = os.path.abspath(destination or (path[-1] + ("." + value.extension if value.extension else "")))
        if value.content is not None:
            with open(target_path, 'wb') as target_file:
                target_file.write(value.content)
        elif value.path != target_path:
            copyfile(value.path, target_path)

    def download_file_set(self, run_uuid: uuid.UUID, path: List[str], destination: Optional[str] = None):
        source_file_set_value: FileSet = self._runs[run_uuid].get(path)

        if destination is None:
            target_file = path[-1] + ".zip"
        elif os.path.isdir(destination):
            target_file = os.path.join(destination, path[-1] + ".zip")
        else:
            target_file = destination

        upload_entries = get_unique_upload_entries(source_file_set_value.file_globs)

        with ZipFile(target_file, 'w') as zipObj:
            for upload_entry in upload_entries:
                zipObj.write(upload_entry.source_path, upload_entry.target_path)

    def get_float_attribute(self, run_uuid: uuid.UUID, path: List[str]) -> FloatAttribute:
        val = self._get_attribute(run_uuid, path, Float)
        return FloatAttribute(val.value)

    def get_int_attribute(self, run_uuid: uuid.UUID, path: List[str]) -> IntAttribute:
        val = self._get_attribute(run_uuid, path, Integer)
        return IntAttribute(val.value)

    def get_bool_attribute(self, run_uuid: uuid.UUID, path: List[str]) -> BoolAttribute:
        val = self._get_attribute(run_uuid, path, Boolean)
        return BoolAttribute(val.value)

    def get_file_attribute(self, run_uuid: uuid.UUID, path: List[str]) -> FileAttribute:
        val = self._get_attribute(run_uuid, path, File)
        return FileAttribute(
            name=os.path.basename(val.path) if val.path else "",
            ext=val.extension or "",
            size=0)

    def get_string_attribute(self, run_uuid: uuid.UUID, path: List[str]) -> StringAttribute:
        val = self._get_attribute(run_uuid, path, String)
        return StringAttribute(val.value)

    def get_datetime_attribute(self, run_uuid: uuid.UUID, path: List[str]) -> DatetimeAttribute:
        val = self._get_attribute(run_uuid, path, Datetime)
        return DatetimeAttribute(val.value)

    def get_float_series_attribute(self, run_uuid: uuid.UUID, path: List[str]) -> FloatSeriesAttribute:
        val = self._get_attribute(run_uuid, path, FloatSeries)
        return FloatSeriesAttribute(val.values[-1] if val.values else None)

    def get_string_series_attribute(self, run_uuid: uuid.UUID, path: List[str]) -> StringSeriesAttribute:
        val = self._get_attribute(run_uuid, path, StringSeries)
        return StringSeriesAttribute(val.values[-1] if val.values else None)

    def get_string_set_attribute(self, run_uuid: uuid.UUID, path: List[str]) -> StringSetAttribute:
        val = self._get_attribute(run_uuid, path, StringSet)
        return StringSetAttribute(set(val.values))

    def _get_attribute(self, run_uuid: uuid.UUID, path: List[str], expected_type: Type[Val]) -> Val:
        if run_uuid not in self._runs:
            raise RunUUIDNotFound(run_uuid)
        value: Optional[Value] = self._runs[run_uuid].get(path)
        str_path = path_to_str(path)
        if value is None:
            raise MetadataInconsistency("Attribute {} not found".format(str_path))
        if isinstance(value, expected_type):
            return value
        raise MetadataInconsistency("Attribute {} is not {}".format(str_path, type.__name__))

    def get_string_series_values(self, run_uuid: uuid.UUID, path: List[str],
                                 offset: int, limit: int) -> StringSeriesValues:
        val = self._get_attribute(run_uuid, path, StringSeries)
        return StringSeriesValues(
            len(val.values),
            [StringPointValue(timestampMillis=-1, step=idx, value=v) for idx, v in enumerate(val.values)]
        )

    def get_float_series_values(self, run_uuid: uuid.UUID, path: List[str],
                                offset: int, limit: int) -> FloatSeriesValues:
        return FloatSeriesValues(0, [])

    def get_image_series_values(self, run_uuid: uuid.UUID, path: List[str],
                                offset: int, limit: int) -> ImageSeriesValues:
        return ImageSeriesValues(0)

    def download_file_series_by_index(self, run_uuid: uuid.UUID, path: List[str],
                                      index: int, destination: str):
        pass

    def get_run_url(self, run_uuid: uuid, workspace: str, project_name: str, short_id: str) -> str:
        return f"offline/{run_uuid}"

    def _get_attribute_values(self, value_dict, path_prefix: List[str]):
        assert isinstance(value_dict, dict)
        for k, value in value_dict.items():
            if isinstance(value, dict):
                yield from self._get_attribute_values(value, path_prefix + [k])
            else:
                attr_type = value.accept(self._attribute_type_converter_value_visitor).value
                attr_path = '/'.join(path_prefix + [k])
                if hasattr(value, "value"):
                    yield attr_path, attr_type, value.value
                else:
                    return attr_path, attr_type, NoValue

    def fetch_atom_attribute_values(self, run_uuid: uuid.UUID, path: List[str]) -> List[Tuple[str, AttributeType, Any]]:
        values = self._get_attribute_values(self._runs[run_uuid].get(path), path)
        namespace_prefix = path_to_str(path)
        if namespace_prefix:
            # don't want to catch "ns/attribute/other" while looking for "ns/attr"
            namespace_prefix += "/"
        return [
            (full_path, attr_type, attr_value)
            for (full_path, attr_type, attr_value) in values
            if full_path.startswith(namespace_prefix)
        ]

    class AttributeTypeConverterValueVisitor(ValueVisitor[AttributeType]):

        def visit_float(self, _: Float) -> AttributeType:
            return AttributeType.FLOAT

        def visit_integer(self, _: Integer) -> AttributeType:
            return AttributeType.INT

        def visit_boolean(self, _: Boolean) -> AttributeType:
            return AttributeType.BOOL

        def visit_string(self, _: String) -> AttributeType:
            return AttributeType.STRING

        def visit_datetime(self, _: Datetime) -> AttributeType:
            return AttributeType.DATETIME

        def visit_file(self, _: File) -> AttributeType:
            return AttributeType.FILE

        def visit_file_set(self, _: FileSet) -> AttributeType:
            return AttributeType.FILE_SET

        def visit_float_series(self, _: FloatSeries) -> AttributeType:
            return AttributeType.FLOAT_SERIES

        def visit_string_series(self, _: StringSeries) -> AttributeType:
            return AttributeType.STRING_SERIES

        def visit_image_series(self, _: FileSeries) -> AttributeType:
            return AttributeType.IMAGE_SERIES

        def visit_string_set(self, _: StringSet) -> AttributeType:
            return AttributeType.STRING_SET

        def visit_git_ref(self, _: GitRef) -> AttributeType:
            return AttributeType.GIT_REF

        def visit_namespace(self, _: Namespace) -> AttributeType:
            raise NotImplementedError

    class NewValueOpVisitor(OperationVisitor[Optional[Value]]):

        def __init__(self, path: List[str], current_value: Optional[Value]):
            self._path = path
            self._current_value = current_value

        def visit_assign_float(self, op: AssignFloat) -> Optional[Value]:
            if self._current_value is not None and not isinstance(self._current_value, Float):
                raise self._create_type_error("assign", Float.__name__)
            return Float(op.value)

        def visit_assign_int(self, op: AssignInt) -> Optional[Value]:
            if self._current_value is not None and not isinstance(self._current_value, Integer):
                raise self._create_type_error("assign", Integer.__name__)
            return Integer(op.value)

        def visit_assign_bool(self, op: AssignBool) -> Optional[Value]:
            if self._current_value is not None and not isinstance(self._current_value, Boolean):
                raise self._create_type_error("assign", Boolean.__name__)
            return Boolean(op.value)

        def visit_assign_string(self, op: AssignString) -> Optional[Value]:
            if self._current_value is not None and not isinstance(self._current_value, String):
                raise self._create_type_error("assign", String.__name__)
            return String(op.value)

        def visit_assign_datetime(self, op: AssignDatetime) -> Optional[Value]:
            if self._current_value is not None and not isinstance(self._current_value, Datetime):
                raise self._create_type_error("assign", Datetime.__name__)
            return Datetime(op.value)

        def visit_upload_file(self, op: UploadFile) -> Optional[Value]:
            if self._current_value is not None and not isinstance(self._current_value, File):
                raise self._create_type_error("save", File.__name__)
            return File(path=op.file_path, extension=op.ext)

        def visit_upload_file_content(self, op: UploadFileContent) -> Optional[Value]:
            if self._current_value is not None and not isinstance(self._current_value, File):
                raise self._create_type_error("upload_files", File.__name__)
            return File(content=base64_decode(op.file_content), extension=op.ext)

        def visit_upload_file_set(self, op: UploadFileSet) -> Optional[Value]:
            if self._current_value is None or op.reset:
                return FileSet(op.file_globs)
            if not isinstance(self._current_value, FileSet):
                raise self._create_type_error("save", FileSet.__name__)
            return FileSet(self._current_value.file_globs + op.file_globs)

        def visit_log_floats(self, op: LogFloats) -> Optional[Value]:
            raw_values = [x.value for x in op.values]
            if self._current_value is None:
                return FloatSeries(raw_values)
            if not isinstance(self._current_value, FloatSeries):
                raise self._create_type_error("log", FloatSeries.__name__)
            return FloatSeries(self._current_value.values + raw_values,
                               min=self._current_value.min,
                               max=self._current_value.max,
                               unit=self._current_value.unit)

        def visit_log_strings(self, op: LogStrings) -> Optional[Value]:
            raw_values = [x.value for x in op.values]
            if self._current_value is None:
                return StringSeries(raw_values)
            if not isinstance(self._current_value, StringSeries):
                raise self._create_type_error("log", StringSeries.__name__)
            return StringSeries(self._current_value.values + raw_values)

        def visit_log_images(self, op: LogImages) -> Optional[Value]:
            raw_values = [File.from_content(base64_decode(x.value.data)) for x in op.values]
            if self._current_value is None:
                return FileSeries(raw_values)
            if not isinstance(self._current_value, FileSeries):
                raise self._create_type_error("log", FileSeries.__name__)
            return FileSeries(self._current_value.values + raw_values)

        def visit_clear_float_log(self, op: ClearFloatLog) -> Optional[Value]:
            # pylint: disable=unused-argument
            if self._current_value is None:
                return FloatSeries([])
            if not isinstance(self._current_value, FloatSeries):
                raise self._create_type_error("clear", FloatSeries.__name__)
            return FloatSeries([],
                               min=self._current_value.min,
                               max=self._current_value.max,
                               unit=self._current_value.unit)

        def visit_clear_string_log(self, op: ClearStringLog) -> Optional[Value]:
            # pylint: disable=unused-argument
            if self._current_value is None:
                return StringSeries([])
            if not isinstance(self._current_value, StringSeries):
                raise self._create_type_error("clear", StringSeries.__name__)
            return StringSeries([])

        def visit_clear_image_log(self, op: ClearImageLog) -> Optional[Value]:
            # pylint: disable=unused-argument
            if self._current_value is None:
                return FileSeries([])
            if not isinstance(self._current_value, FileSeries):
                raise self._create_type_error("clear", FileSeries.__name__)
            return FileSeries([])

        def visit_config_float_series(self, op: ConfigFloatSeries) -> Optional[Value]:
            if self._current_value is None:
                return FloatSeries([], min=op.min, max=op.max, unit=op.unit)
            if not isinstance(self._current_value, FloatSeries):
                raise self._create_type_error("log", FloatSeries.__name__)
            return FloatSeries(self._current_value.values, min=op.min, max=op.max, unit=op.unit)

        def visit_add_strings(self, op: AddStrings) -> Optional[Value]:
            if self._current_value is None:
                return StringSet(op.values)
            if not isinstance(self._current_value, StringSet):
                raise self._create_type_error("add", StringSet.__name__)
            return StringSet(self._current_value.values.union(op.values))

        def visit_remove_strings(self, op: RemoveStrings) -> Optional[Value]:
            if self._current_value is None:
                return StringSet(set())
            if not isinstance(self._current_value, StringSet):
                raise self._create_type_error("remove", StringSet.__name__)
            return StringSet(self._current_value.values.difference(op.values))

        def visit_clear_string_set(self, op: ClearStringSet) -> Optional[Value]:
            # pylint: disable=unused-argument
            if self._current_value is None:
                return StringSet(set())
            if not isinstance(self._current_value, StringSet):
                raise self._create_type_error("clear", StringSet.__name__)
            return StringSet(set())

        def visit_delete_files(self, op: DeleteFiles) -> Optional[Value]:
            # pylint: disable=unused-argument
            if self._current_value is None:
                return FileSet([])
            if not isinstance(self._current_value, FileSet):
                raise self._create_type_error("delete_files", FileSet.__name__)
            # It is not important to support deleting properly in debug mode, let's just ignore this operation
            return self._current_value

        def visit_delete_attribute(self, op: DeleteAttribute) -> Optional[Value]:
            # pylint: disable=unused-argument
            if self._current_value is None:
                raise MetadataInconsistency(
                    "Cannot perform delete operation on {}. Attribute is undefined.".format(self._path))
            return None

        def _create_type_error(self, op_name, expected):
            return MetadataInconsistency("Cannot perform {} operation on {}. Expected {}, {} found."
                                         .format(op_name, self._path, expected, type(self._current_value)))



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging
import threading
import time
from collections import namedtuple
from itertools import groupby
from queue import Queue, Empty

from bravado.exception import HTTPUnprocessableEntity

from neptune.exceptions import NeptuneException
from neptune.internal.channels.channels import ChannelIdWithValues, ChannelValue, \
    ChannelType, ChannelNamespace
from neptune.internal.threads.neptune_thread import NeptuneThread

_logger = logging.getLogger(__name__)


class ChannelsValuesSender(object):
    _QUEUED_CHANNEL_VALUE = namedtuple(
        "QueuedChannelValue",
        ['channel_id', 'channel_name', 'channel_type', 'channel_value', 'channel_namespace']
    )

    __LOCK = threading.RLock()

    def __init__(self, experiment):
        self._experiment = experiment
        self._values_queue = None
        self._sending_thread = None
        self._user_channel_name_to_id_map = dict()
        self._system_channel_name_to_id_map = dict()

    # pylint:disable=protected-access
    def send(self, channel_name, channel_type, channel_value, channel_namespace=ChannelNamespace.USER):
        # Before taking the lock here, we need to check if the sending thread is not running yet.
        # Otherwise, the sending thread could call send() while being join()-ed, which would result
        # in a deadlock.
        if not self._is_running():
            with self.__LOCK:
                if not self._is_running():
                    self._start()

        if channel_namespace == ChannelNamespace.USER:
            namespaced_channel_map = self._user_channel_name_to_id_map
        else:
            namespaced_channel_map = self._system_channel_name_to_id_map

        if channel_name in namespaced_channel_map:
            channel_id = namespaced_channel_map[channel_name]
        else:
            response = self._experiment._create_channel(channel_name, channel_type, channel_namespace)
            channel_id = response.id
            namespaced_channel_map[channel_name] = channel_id

        self._values_queue.put(self._QUEUED_CHANNEL_VALUE(
            channel_id=channel_id,
            channel_name=channel_name,
            channel_type=channel_type,
            channel_value=channel_value,
            channel_namespace=channel_namespace
        ))

    def join(self):
        with self.__LOCK:
            if self._is_running():
                self._sending_thread.interrupt()
                self._sending_thread.join()
                self._sending_thread = None
                self._values_queue = None

    def _is_running(self):
        return self._values_queue is not None and self._sending_thread is not None and self._sending_thread.is_alive()

    def _start(self):
        self._values_queue = Queue()
        self._sending_thread = ChannelsValuesSendingThread(self._experiment, self._values_queue)
        self._sending_thread.start()


class ChannelsValuesSendingThread(NeptuneThread):
    _SLEEP_TIME = 5
    _MAX_VALUES_BATCH_LENGTH = 100
    _MAX_IMAGE_VALUES_BATCH_SIZE = 10485760  # 10 MB

    def __init__(self, experiment, values_queue):
        super(ChannelsValuesSendingThread, self).__init__(is_daemon=False)
        self._values_queue = values_queue
        self._experiment = experiment
        self._sleep_time = self._SLEEP_TIME
        self._values_batch = []

    def run(self):
        while self.should_continue_running() or not self._values_queue.empty():
            try:
                sleep_start = time.time()
                self._values_batch.append(self._values_queue.get(timeout=max(self._sleep_time, 0)))
                self._values_queue.task_done()
                self._sleep_time -= time.time() - sleep_start
            except Empty:
                self._sleep_time = 0

            image_values_batch_size = sum([len(v.channel_value.y['image_value']['data'] or [])
                                           for v in self._values_batch
                                           if v.channel_type == ChannelType.IMAGE.value])
            if self._sleep_time <= 0 \
                    or len(self._values_batch) >= self._MAX_VALUES_BATCH_LENGTH \
                    or image_values_batch_size >= self._MAX_IMAGE_VALUES_BATCH_SIZE:  # pylint:disable=line-too-long
                self._process_batch()

        self._process_batch()

    def _process_batch(self):
        send_start = time.time()
        if self._values_batch:
            try:
                self._send_values(self._values_batch)
                self._values_batch = []
            except (NeptuneException, IOError):
                _logger.exception('Failed to send channel value.')
        self._sleep_time = self._SLEEP_TIME - (time.time() - send_start)

    def _send_values(self, queued_channels_values):
        def get_channel_metadata(value):
            return value.channel_id, value.channel_name, value.channel_type, value.channel_namespace

        queued_grouped_by_channel = {channel_metadata: list(values)
                                     for channel_metadata, values
                                     in groupby(sorted(queued_channels_values, key=get_channel_metadata),
                                                get_channel_metadata)}
        channels_with_values = []
        for channel_metadata in queued_grouped_by_channel:
            channel_values = []
            for queued_value in queued_grouped_by_channel[channel_metadata]:
                channel_values.append(ChannelValue(ts=queued_value.channel_value.ts,
                                                   x=queued_value.channel_value.x,
                                                   y=queued_value.channel_value.y))
            channels_with_values.append(
                ChannelIdWithValues(*channel_metadata, channel_values))

        try:
            # pylint:disable=protected-access
            self._experiment._send_channels_values(channels_with_values)
        except HTTPUnprocessableEntity as e:
            message = "Maximum storage limit reached"
            try:
                message = e.response.json()["message"]
            finally:
                _logger.warning('Failed to send channel value: %s', message)
        except (NeptuneException, IOError):
            _logger.exception('Failed to send channel value.')



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import unittest
import uuid
from tempfile import NamedTemporaryFile, TemporaryDirectory

import mock
from mock import MagicMock, patch

from neptune.new.internal.backends.hosted_file_operations import upload_file_attribute, upload_file_set_attribute, \
    download_file_attribute, _get_content_disposition_filename, _attribute_upload_response_handler, \
    download_file_set_attribute
from neptune.utils import IS_WINDOWS


class TestHostedFileOperations(unittest.TestCase):
    # pylint:disable=protected-access

    @unittest.skipIf(IS_WINDOWS, "Windows behaves strangely")
    @patch('neptune.new.internal.backends.hosted_file_operations._upload_loop')
    def test_upload_file_attribute(self, upload_loop_mock):
        # given
        exp_uuid = uuid.uuid4()
        swagger_mock = self._get_swagger_mock()
        upload_loop_mock.return_value = b'null'

        # when
        with NamedTemporaryFile("w") as f:
            upload_file_attribute(
                swagger_client=swagger_mock,
                run_uuid=exp_uuid,
                attribute="target/path.txt",
                source=f.name,
                ext="txt")

        # then
        upload_loop_mock.assert_called_once_with(
            file_chunk_stream=mock.ANY,
            response_handler=_attribute_upload_response_handler,
            http_client=swagger_mock.swagger_spec.http_client,
            url="ui.neptune.ai/attributes/upload",
            query_params={
                "experimentId": str(exp_uuid),
                "attribute": "target/path.txt",
                "ext": "txt"
            })

    @unittest.skipIf(IS_WINDOWS, "Windows behaves strangely")
    @patch('neptune.new.internal.backends.hosted_file_operations._upload_loop')
    def test_upload_file_attribute_from_stream(self, upload_loop_mock):
        # given
        exp_uuid = uuid.uuid4()
        swagger_mock = self._get_swagger_mock()
        upload_loop_mock.return_value = b'null'

        # when
        upload_file_attribute(
            swagger_client=swagger_mock,
            run_uuid=exp_uuid,
            attribute="target/path.txt",
            source=b"Some content of test stream",
            ext="txt")

        # then
        upload_loop_mock.assert_called_once_with(
            file_chunk_stream=mock.ANY,
            response_handler=_attribute_upload_response_handler,
            http_client=swagger_mock.swagger_spec.http_client,
            url="ui.neptune.ai/attributes/upload",
            query_params={
                "experimentId": str(exp_uuid),
                "attribute": "target/path.txt",
                "ext": "txt"
            })

    @unittest.skipIf(IS_WINDOWS, "Windows behaves strangely")
    @patch('neptune.new.internal.backends.hosted_file_operations._upload_loop')
    @patch('neptune.new.internal.utils.glob', new=lambda path, recursive=False: [path.replace('*', 'file.txt')])
    def test_upload_single_file_in_file_set_attribute(self, upload_loop_mock):
        # given
        exp_uuid = uuid.uuid4()
        swagger_mock = self._get_swagger_mock()
        upload_loop_mock.return_value = b'null'

        # when
        with NamedTemporaryFile("w") as temp_file:
            upload_file_set_attribute(
                swagger_client=swagger_mock,
                run_uuid=exp_uuid,
                attribute="some/attribute",
                file_globs=[temp_file.name],
                reset=True)

        # then
        upload_loop_mock.assert_called_once_with(
            file_chunk_stream=mock.ANY,
            response_handler=_attribute_upload_response_handler,
            http_client=swagger_mock.swagger_spec.http_client,
            url="ui.neptune.ai/uploadFileSetChunk",
            query_params={
                "experimentId": str(exp_uuid),
                "attribute": "some/attribute",
                "reset": "True",
                "path": os.path.basename(temp_file.name)
            })

    @unittest.skipIf(IS_WINDOWS, "Windows behaves strangely")
    @patch('neptune.new.internal.backends.hosted_file_operations.upload_raw_data')
    @patch('neptune.new.internal.utils.glob', new=lambda path, recursive=False: [path.replace('*', 'file.txt')])
    def test_upload_multiple_files_in_file_set_attribute(self, upload_raw_data_mock):
        # given
        exp_uuid = uuid.uuid4()
        swagger_mock = self._get_swagger_mock()
        upload_raw_data_mock.return_value = b'null'

        # when
        with NamedTemporaryFile("w") as temp_file_1:
            with NamedTemporaryFile("w") as temp_file_2:
                upload_file_set_attribute(
                    swagger_client=swagger_mock,
                    run_uuid=exp_uuid,
                    attribute="some/attribute",
                    file_globs=[temp_file_1.name, temp_file_2.name],
                    reset=True)

        # then
        upload_raw_data_mock.assert_called_once_with(
            http_client=swagger_mock.swagger_spec.http_client,
            url="ui.neptune.ai/uploadFileSetTar",
            data=mock.ANY,
            headers={"Content-Type": "application/octet-stream"},
            query_params={
                "experimentId": str(exp_uuid),
                "attribute": "some/attribute",
                "reset": "True"
            })

    @unittest.skipIf(IS_WINDOWS, "Windows behaves strangely")
    @patch('neptune.new.internal.backends.hosted_file_operations.upload_raw_data')
    def test_missing_files_or_directory(self, upload_raw_data_mock):
        # given
        exp_uuid = uuid.uuid4()
        swagger_mock = self._get_swagger_mock()
        upload_raw_data_mock.return_value = b'null'

        # when
        with NamedTemporaryFile("w") as temp_file_1:
            with NamedTemporaryFile("w") as temp_file_2:
                with TemporaryDirectory() as temp_dir:
                    upload_file_set_attribute(
                        swagger_client=swagger_mock,
                        run_uuid=exp_uuid,
                        attribute="some/attribute",
                        file_globs=[temp_file_1.name, temp_file_2.name, os.path.abspath("missing_file"), temp_dir],
                        reset=True)

        # then
        upload_raw_data_mock.assert_called_once_with(
            http_client=swagger_mock.swagger_spec.http_client,
            url="ui.neptune.ai/uploadFileSetTar",
            data=mock.ANY,
            headers={"Content-Type": "application/octet-stream"},
            query_params={
                "experimentId": str(exp_uuid),
                "attribute": "some/attribute",
                "reset": "True"
            })

    def test_get_content_disposition_filename(self):
        # given
        response_mock = MagicMock()
        response_mock.headers = {'Content-Disposition': 'attachment; filename="sample.file"'}

        # when
        filename = _get_content_disposition_filename(response_mock)

        # then
        self.assertEqual(filename, "sample.file")

    @patch('neptune.new.internal.backends.hosted_file_operations._store_response_as_file')
    @patch('neptune.new.internal.backends.hosted_file_operations._download_raw_data')
    def test_download_file_attribute(self, download_raw, store_response_mock):
        # given
        swagger_mock = self._get_swagger_mock()
        exp_uuid = uuid.uuid4()

        # when
        download_file_attribute(swagger_client=swagger_mock,
                                run_uuid=exp_uuid,
                                attribute="some/attribute",
                                destination=None)

        # then
        download_raw.assert_called_once_with(
            http_client=swagger_mock.swagger_spec.http_client,
            url="ui.neptune.ai/attributes/download",
            headers={"Accept": "application/octet-stream"},
            query_params={
                "experimentId": str(exp_uuid),
                "attribute": "some/attribute"
            },
        )
        store_response_mock.assert_called_once_with(download_raw.return_value, None)

    @patch('neptune.new.internal.backends.hosted_file_operations._store_response_as_file')
    @patch('neptune.new.internal.backends.hosted_file_operations._download_raw_data')
    @patch('neptune.new.internal.backends.hosted_file_operations._get_download_url', new=lambda _, _id: "some_url")
    def test_download_file_set_attribute(self, download_raw, store_response_mock):
        # given
        swagger_mock = self._get_swagger_mock()
        download_id = uuid.uuid4()

        # when
        download_file_set_attribute(swagger_client=swagger_mock,
                                    download_id=download_id,
                                    destination=None)

        # then
        download_raw.assert_called_once_with(
            http_client=swagger_mock.swagger_spec.http_client,
            url="some_url",
            headers={"Accept": "application/zip"},
        )
        store_response_mock.assert_called_once_with(download_raw.return_value, None)

    @staticmethod
    def _get_swagger_mock():
        swagger_mock = MagicMock()
        swagger_mock.swagger_spec.http_client = MagicMock()
        swagger_mock.swagger_spec.api_url = "ui.neptune.ai"
        swagger_mock.api.uploadFileSetAttributeChunk.operation.path_name = "/uploadFileSetChunk"
        swagger_mock.api.uploadFileSetAttributeTar.operation.path_name = "/uploadFileSetTar"
        swagger_mock.api.uploadPath.operation.path_name = "/uploadPath"
        swagger_mock.api.uploadAttribute.operation.path_name = "/attributes/upload"
        swagger_mock.api.downloadAttribute.operation.path_name = "/attributes/download"
        swagger_mock.api.downloadFileSetAttributeZip.operation.path_name = "/attributes/downloadFileSetZip"
        swagger_mock.api.downloadFileSetAttributeZip.operation.path_name = "/attributes/downloadFileSetZip"
        swagger_mock.api.download.operation.path_name = "/download"
        return swagger_mock


if __name__ == '__main__':
    unittest.main()



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import List, TYPE_CHECKING

from neptune.new.attributes.atoms.datetime import Datetime as DatetimeAttr
from neptune.new.attributes.atoms.file import File as FileAttr
from neptune.new.attributes.file_set import FileSet as FileSetAttr
from neptune.new.attributes.atoms.float import Float as FloatAttr
from neptune.new.attributes.atoms.integer import Integer as IntegerAttr
from neptune.new.attributes.atoms.boolean import Boolean as BooleanAttr
from neptune.new.attributes.atoms.string import String as StringAttr
from neptune.new.attributes.namespace import Namespace as NamespaceAttr
from neptune.new.attributes.attribute import Attribute
from neptune.new.attributes.series.float_series import FloatSeries as FloatSeriesAttr
from neptune.new.attributes.series.file_series import FileSeries as ImageSeriesAttr
from neptune.new.attributes.series.string_series import StringSeries as StringSeriesAttr
from neptune.new.attributes.sets.string_set import StringSet as StringSetAttr
from neptune.new.exceptions import OperationNotSupported
from neptune.new.types import Boolean, Integer
from neptune.new.types.atoms import GitRef
from neptune.new.types.atoms.datetime import Datetime
from neptune.new.types.atoms.file import File
from neptune.new.types.atoms.float import Float
from neptune.new.types.atoms.string import String
from neptune.new.types.file_set import FileSet
from neptune.new.types.namespace import Namespace
from neptune.new.types.series.float_series import FloatSeries
from neptune.new.types.series.file_series import FileSeries
from neptune.new.types.series.string_series import StringSeries
from neptune.new.types.sets.string_set import StringSet
from neptune.new.types.value_visitor import ValueVisitor

if TYPE_CHECKING:
    from neptune.new import Run


class ValueToAttributeVisitor(ValueVisitor[Attribute]):

    def __init__(self, run: 'Run', path: List[str]):
        self._run = run
        self._path = path

    def visit_float(self, _: Float) -> Attribute:
        return FloatAttr(self._run, self._path)

    def visit_integer(self, _: Integer) -> Attribute:
        return IntegerAttr(self._run, self._path)

    def visit_boolean(self, _: Boolean) -> Attribute:
        return BooleanAttr(self._run, self._path)

    def visit_string(self, _: String) -> Attribute:
        return StringAttr(self._run, self._path)

    def visit_datetime(self, _: Datetime) -> Attribute:
        return DatetimeAttr(self._run, self._path)

    def visit_file(self, _: File) -> Attribute:
        return FileAttr(self._run, self._path)

    def visit_file_set(self, _: FileSet) -> Attribute:
        return FileSetAttr(self._run, self._path)

    def visit_float_series(self, _: FloatSeries) -> Attribute:
        return FloatSeriesAttr(self._run, self._path)

    def visit_string_series(self, _: StringSeries) -> Attribute:
        return StringSeriesAttr(self._run, self._path)

    def visit_image_series(self, _: FileSeries) -> Attribute:
        return ImageSeriesAttr(self._run, self._path)

    def visit_string_set(self, _: StringSet) -> Attribute:
        return StringSetAttr(self._run, self._path)

    def visit_git_ref(self, _: GitRef) -> Attribute:
        raise OperationNotSupported("Cannot create custom attribute of type GitRef")

    def visit_namespace(self, _: Namespace) -> Attribute:
        return NamespaceAttr(self._run, self._path)



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
from contextlib import contextmanager
from datetime import datetime
import abc


class ClientFeatures(abc.ABC):
    params = {
        'init_text_parameter': 'some text',
        'init_number parameter': 42,
        'init_list': [1, 2, 3],
        'init_datetime': datetime.now()
    }
    properties = {
        'init_text_property': 'some text',
        'init_number property': 42,
        'init_list': [1, 2, 3],
        'init_datetime': datetime.now(),
    }

    data_dir = 'alpha_integration_dev/data'
    img_path = f'{data_dir}/g.png'
    text_file_path = f'{data_dir}/text.txt'

    @contextmanager
    def with_assert_raises(self, exception):
        """Checks given exception is raised inside block,
        raises AssertionError otherwise"""
        try:
            yield
        except exception:
            pass
        else:
            raise AssertionError(f'Should raise {exception}')

    @staticmethod
    def _remove_file_if_exists(filepath):
        try:
            os.remove(filepath)
        except OSError:
            pass

    @contextmanager
    def with_check_if_file_appears(self, filepath):
        """Checks if file will be present when leaving the block.
        File is removed if exists when entering the block."""
        self._remove_file_if_exists(filepath)

        try:
            yield
            assert os.path.exists(filepath)
        except:
            raise
        else:
            self._remove_file_if_exists(filepath)

    @abc.abstractmethod
    def modify_tags(self):
        pass

    @abc.abstractmethod
    def modify_properties(self):
        pass

    @abc.abstractmethod
    def log_std(self):
        """system streams / monitoring logs"""

    @abc.abstractmethod
    def log_series(self):
        pass

    @abc.abstractmethod
    def handle_files_and_images(self):
        pass

    @abc.abstractmethod
    def handle_directories(self):
        pass

    @abc.abstractmethod
    def finalize(self):
        pass

    def run(self):
        self.modify_tags()
        self.modify_properties()
        self.log_std()
        self.log_series()
        self.handle_files_and_images()
        self.handle_directories()

        self.finalize()



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import uuid
from typing import Union, Optional, Iterable

from neptune.new.internal.backends.hosted_neptune_backend import HostedNeptuneBackend
from neptune.new.internal.utils import verify_type, verify_collection_type
from neptune.new.runs_table import RunsTable


class Project:
    """A class for managing a Neptune project and retrieving information from it.

    You may also want to check `Project docs page`_.

    .. _Project docs page:
       https://docs.neptune.ai/api-reference/project
    """

    def __init__(self,
                 _uuid: uuid.UUID,
                 backend: HostedNeptuneBackend):
        self._uuid = _uuid
        self._backend = backend

    # pylint:disable=redefined-builtin
    def fetch_runs_table(self,
                         id: Optional[Union[str, Iterable[str]]] = None,
                         state: Optional[Union[str, Iterable[str]]] = None,
                         owner: Optional[Union[str, Iterable[str]]] = None,
                         tag: Optional[Union[str, Iterable[str]]] = None
                         ) -> RunsTable:
        """Retrieve runs matching the specified criteria.

        All parameters are optional, each of them specifies a single criterion.
        Only runs matching all of the criteria will be returned.

        Args:
            id (str or list of str, optional): A run's id or list of ids.
                E.g. `'SAN-1'` or `['SAN-1', 'SAN-2']`.
                Matching any element of the list is sufficient to pass the criterion.
                Defaults to `None`.
            state (str or list of str, optional): A run's state like or list of states.
                E.g. `'running'` or `['idle', 'running']`.
                Possible values: 'idle', 'running'.
                Defaults to `None`.
                Matching any element of the list is sufficient to pass the criterion.
            owner (str or list of str, optional): Username of the run's owner  or a list of owners.
                E.g. 'josh' or ['frederic', 'josh'].
                The user who created the tracked run is an owner.
                Defaults to `None`.
                Matching any element of the list is sufficient to pass the criterion.
            tag (str or list of str, optional): An experiment tag or list of tags.
                E.g. `'lightGBM'` or ['pytorch', 'cycleLR'].
                Defaults to `None`.
                Only experiments that have all specified tags will match this criterion.

        Returns:
            ``RunsTable``: object containing experiments matching the specified criteria.

            Use `.to_pandas()` to convert it to Pandas `DataFrame`.

        Examples:
            >>> import neptune.new as neptune

            >>> # Fetch project 'jackie/sandbox'
            ... project = neptune.get_project(name='jackie/sandbox')

            >>> # Fetch all Runs metadata as Pandas DataFrame
            ... runs_table_df = project.fetch_runs_table().to_pandas()

            >>> # Sort runs by creation time
            ... runs_table_df = runs_table_df.sort_values(by='sys/creation_time', ascending=False)

            >>> # Extract the last runs id
            ... last_run_id = runs_table_df['sys/id'].values[0]

            You can also filter the runs table by state, owner or tag or a combination:

            >>> # Fetch only inactive runs
            ... runs_table_df = project.fetch_runs_table(state='idle').to_pandas()

            >>> # Fetch only runs created by CI service
            ... runs_table_df = project.fetch_runs_table(owner='my_company_ci_service').to_pandas()

            >>> # Fetch only runs that have both 'Exploration' and 'Optuna' tag
            ... runs_table_df = project.fetch_runs_table(tag=['Exploration', 'Optuna']).to_pandas()

            >>> # You can combine conditions. Runs satisfying all conditions will be fetched
            ... runs_table_df = project.fetch_runs_table(state='idle', tag='Exploration').to_pandas()

        You may also want to check `fetch_runs_table docs page`_.

        .. _fetch_runs_table docs page:
            https://docs.neptune.ai/api-reference/project#fetch_runs_table
        """
        id = self._as_list("id", id)
        state = self._as_list("state", state)
        owner = self._as_list("owner", owner)
        tags = self._as_list("tag", tag)

        leaderboard_entries = self._backend.get_leaderboard(self._uuid, id, state, owner, tags)

        return RunsTable(self._backend, leaderboard_entries)

    @staticmethod
    def _as_list(name: str, value: Optional[Union[str, Iterable[str]]]) -> Optional[Iterable[str]]:
        verify_type(name, value, (type(None), str, Iterable))
        if value is None:
            return None
        if isinstance(value, str):
            return [value]
        verify_collection_type(name, value, str)
        return value



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Optional

from neptune.new.internal.containers.storage_queue import StorageQueue
from neptune.new.internal.operation import Operation
from neptune.new.internal.operation_processors.operation_processor import OperationProcessor


class OfflineOperationProcessor(OperationProcessor):

    def __init__(self, queue: StorageQueue[Operation]):
        self._queue = queue

    def enqueue_operation(self, op: Operation, wait: bool) -> None:
        # pylint: disable=unused-argument
        self._queue.put(op)

    def wait(self):
        self.flush()

    def flush(self):
        self._queue.flush()

    def start(self):
        pass

    def stop(self, seconds: Optional[float] = None):
        pass



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

try:
    import psutil
    PSUTIL_INSTALLED = True
except ImportError:
    PSUTIL_INSTALLED = False


class SystemMonitor(object):
    @staticmethod
    def requirements_installed():
        return PSUTIL_INSTALLED

    @staticmethod
    def cpu_count():
        return psutil.cpu_count()

    @staticmethod
    def cpu_percent():
        return psutil.cpu_percent()

    @staticmethod
    def virtual_memory():
        return psutil.virtual_memory()



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging

from typing import TYPE_CHECKING, Optional

from neptune.new.internal.background_job import BackgroundJob
from neptune.new.internal.threading.daemon import Daemon

if TYPE_CHECKING:
    from neptune.new.run import Run

_logger = logging.getLogger(__name__)


class PingBackgroundJob(BackgroundJob):

    def __init__(self, period: float = 10):
        self._period = period
        self._thread = None
        self._started = False

    def start(self, run: 'Run'):
        self._thread = self.ReportingThread(self._period, run)
        self._thread.start()
        self._started = True

    def stop(self):
        if not self._started:
            return
        self._thread.interrupt()

    def join(self, seconds: Optional[float] = None):
        if not self._started:
            return
        self._thread.join(seconds)

    class ReportingThread(Daemon):

        def __init__(
                self,
                period: float,
                run: 'Run'):
            super().__init__(sleep_time=period)
            self._run = run

        @Daemon.ConnectionRetryWrapper(
            kill_message=(
                "Killing Neptune ping thread. Your run's status will not be updated and"
                " the run will be shown as inactive."
            )
        )
        def work(self) -> None:
            self._run.ping()



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import json
import logging
import os
from glob import glob
from pathlib import Path
from threading import Event
from typing import TypeVar, List, Callable, Optional, Tuple

from neptune.new.exceptions import MalformedOperation
from neptune.new.internal.containers.storage_queue import StorageQueue
from neptune.new.internal.utils.json_file_splitter import JsonFileSplitter
from neptune.new.internal.utils.sync_offset_file import SyncOffsetFile

T = TypeVar('T')

_logger = logging.getLogger(__name__)


class DiskQueue(StorageQueue[T]):

    # NOTICE: This class is thread-safe as long as there is only one consumer and one producer.

    def __init__(
            self,
            dir_path: Path,
            to_dict: Callable[[T], dict],
            from_dict: Callable[[dict], T],
            max_file_size: int = 64 * 1024**2):
        self._dir_path = dir_path.resolve()
        self._to_dict = to_dict
        self._from_dict = from_dict
        self._max_file_size = max_file_size

        try:
            os.makedirs(self._dir_path)
        except FileExistsError:
            pass

        self._last_ack_file = SyncOffsetFile(dir_path / "last_ack_version", default=0)
        self._last_put_file = SyncOffsetFile(dir_path / "last_put_version", default=0)

        self._read_file_version, self._write_file_version = self._get_first_and_last_log_file_version()
        self._writer = open(self._get_log_file(self._write_file_version), "a")
        self._reader = JsonFileSplitter(self._get_log_file(self._read_file_version))
        self._file_size = 0
        self._should_skip_to_ack = True

        self._event_empty = Event()
        if self.is_empty():
            self._event_empty.set()
        else:
            self._event_empty.clear()

    def put(self, obj: T) -> int:
        version = self._last_put_file.read_local() + 1
        self._event_empty.clear()
        _json = json.dumps(self._serialize(obj, version))
        if self._file_size + len(_json) > self._max_file_size:
            self._writer.flush()
            self._writer.close()
            self._writer = open(self._get_log_file(version), "a")
            self._file_size = 0
            self._write_file_version = version
        self._writer.write(_json + "\n")
        self._last_put_file.write(version)
        self._file_size += len(_json) + 1
        return version

    def get(self) -> Tuple[Optional[T], int]:
        if self._should_skip_to_ack:
            return self._skip_and_get()
        else:
            return self._get()

    def _skip_and_get(self) -> Tuple[Optional[T], int]:
        ack_version = self._last_ack_file.read_local()
        ver = -1
        while True:
            obj, next_ver = self._get()
            if obj is None:
                return None, ver
            ver = next_ver
            if ver > ack_version:
                self._should_skip_to_ack = False
                if ver > ack_version + 1:
                    _logger.warning("Possible data loss. Last acknowledged operation version: %d, next: %d",
                                    ack_version, ver)
                return obj, ver

    def _get(self) -> Tuple[Optional[T], int]:
        _json = self._reader.get()
        if not _json:
            if self._read_file_version >= self._write_file_version:
                return None, -1
            self._reader.close()
            self._read_file_version = self._next_log_file_version(self._read_file_version)
            self._reader = JsonFileSplitter(self._get_log_file(self._read_file_version))
            # It is safe. Max recursion level is 2.
            return self._get()
        try:
            return self._deserialize(_json)
        except Exception as e:
            raise MalformedOperation from e

    def get_batch(self, size: int) -> Tuple[List[T], int]:
        first, ver = self.get()
        if not first:
            return [], ver
        ret = [first]
        for _ in range(0, size - 1):
            obj, next_ver = self._get()
            if not obj:
                break
            ver = next_ver
            ret.append(obj)
        return ret, ver

    def flush(self):
        self._writer.flush()
        self._last_ack_file.flush()
        self._last_put_file.flush()

    def close(self):
        self._reader.close()
        self._writer.close()
        self._last_ack_file.close()
        self._last_put_file.close()

    def wait_for_empty(self, seconds: Optional[float] = None) -> None:
        self._event_empty.wait(seconds)

    def ack(self, version: int) -> None:
        self._last_ack_file.write(version)
        if self.is_empty():
            self._event_empty.set()

        log_versions = self._get_all_log_file_versions()
        for i in range(0, len(log_versions) - 1):
            if log_versions[i + 1] <= version:
                os.remove(self._get_log_file(log_versions[i]))
            else:
                break

    def is_empty(self) -> bool:
        return self.size() == 0

    def size(self) -> int:
        return self._last_put_file.read_local() - self._last_ack_file.read_local()

    def _get_log_file(self, index: int) -> str:
        return "{}/data-{}.log".format(self._dir_path, index)

    def _get_all_log_file_versions(self):
        log_files = glob("{}/data-*.log".format(self._dir_path))
        if not log_files:
            return 1, 1
        return sorted([int(file[len(str(self._dir_path)) + 6:-4]) for file in log_files])

    def _get_first_and_last_log_file_version(self) -> (int, int):
        log_versions = self._get_all_log_file_versions()
        return min(log_versions), max(log_versions)

    def _next_log_file_version(self, version: int) -> int:
        log_versions = self._get_all_log_file_versions()
        for i, val in enumerate(log_versions):
            if val == version:
                return log_versions[i + 1]
        raise ValueError("Missing log file with version > {}".format(version))

    def _serialize(self, obj: T, version: int) -> dict:
        return {
            "obj": self._to_dict(obj),
            "version": version
        }

    def _deserialize(self, data: dict) -> Tuple[T, int]:
        return self._from_dict(data["obj"]), data["version"]



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from neptune.internal.hardware.cgroup.cgroup_monitor import CGroupMonitor
from neptune.internal.hardware.gauges.gauge_mode import GaugeMode
from neptune.internal.hardware.resources.gpu_card_indices_provider import GPUCardIndicesProvider
from neptune.internal.hardware.resources.system_resource_info import SystemResourceInfo


class SystemResourceInfoFactory(object):
    def __init__(self, system_monitor, gpu_monitor, os_environ):
        self.__system_monitor = system_monitor
        self.__gpu_monitor = gpu_monitor
        self.__gpu_card_indices_provider = GPUCardIndicesProvider(
            cuda_visible_devices=os_environ.get(u'CUDA_VISIBLE_DEVICES'),
            gpu_card_count=self.__gpu_monitor.get_card_count())

    def create(self, gauge_mode):
        if gauge_mode == GaugeMode.SYSTEM:
            return self.__create_whole_system_resource_info()
        elif gauge_mode == GaugeMode.CGROUP:
            return self.__create_cgroup_resource_info()
        else:
            raise ValueError(str(u'Unknown gauge mode: {}'.format(gauge_mode)))

    def __create_whole_system_resource_info(self):
        return SystemResourceInfo(
            cpu_core_count=float(self.__system_monitor.cpu_count()),
            memory_amount_bytes=self.__system_monitor.virtual_memory().total,
            gpu_card_indices=self.__gpu_card_indices_provider.get(),
            gpu_memory_amount_bytes=self.__gpu_monitor.get_top_card_memory_in_bytes()
        )

    def __create_cgroup_resource_info(self):
        cgroup_monitor = CGroupMonitor.create()

        return SystemResourceInfo(
            cpu_core_count=cgroup_monitor.get_cpu_usage_limit_in_cores(),
            memory_amount_bytes=cgroup_monitor.get_memory_limit_in_bytes(),
            gpu_card_indices=self.__gpu_card_indices_provider.get(),
            gpu_memory_amount_bytes=self.__gpu_monitor.get_top_card_memory_in_bytes()
        )



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import TypeVar, TYPE_CHECKING, List

import click

from neptune.new.types import File

from neptune.new.internal.utils import is_collection

from neptune.new.types.series.series import Series

if TYPE_CHECKING:
    from neptune.new.types.value_visitor import ValueVisitor

Ret = TypeVar('Ret')


class FileSeries(Series):

    def __init__(self, values, **kwargs):
        if not is_collection(values):
            raise TypeError("`values` is not a collection")
        self._values = [File.create_from(value) for value in values]

        self.name = kwargs.pop("name", None)
        self.description = kwargs.pop("description", None)
        if kwargs:
            click.echo("Warning: unexpected arguments ({kwargs}) in FileSeries".format(kwargs=kwargs), err=True)

    def accept(self, visitor: 'ValueVisitor[Ret]') -> Ret:
        return visitor.visit_image_series(self)

    @property
    def values(self) -> List[File]:
        return self._values

    def __str__(self):
        return "FileSeries({})".format(str(self.values))



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from random import randint, uniform

from mock import MagicMock

from tests.neptune.random_utils import a_string, a_uuid_string, a_timestamp


def a_project():
    project = MagicMock()
    project.id = a_uuid_string()
    project.name = a_string()
    return project


def an_invited_project_member():
    invitation_info = MagicMock()
    invitation_info.id = a_uuid_string()
    invitation_info.email = a_string() + '@example.com'

    project_member = MagicMock()
    project_member.invitationInfo = invitation_info
    project_member.registeredMemberInfo = None
    project_member.role = 'member'

    return project_member


def a_registered_project_member(username=None):
    if username is None:
        username = a_string()

    registered_member_info = MagicMock()
    registered_member_info.avatarSource = 'default'
    registered_member_info.avatarUrl = ''
    registered_member_info.firstName = a_string()
    registered_member_info.lastName = a_string()
    registered_member_info.username = username

    project_member = MagicMock()
    project_member.invitationInfo = None
    project_member.registeredMemberInfo = registered_member_info
    project_member.role = 'manager'

    return project_member


def an_experiment_states(creating=None, waiting=None, initializing=None, running=None, cleaning=None, aborted=None,
                         crashed=None, failed=None, succeeded=None, not_responding=None):
    def random_state_count():
        return randint(1, 100)

    experiment_states = MagicMock()
    experiment_states.creating = creating or random_state_count()
    experiment_states.waiting = waiting or random_state_count()
    experiment_states.initializing = initializing or random_state_count()
    experiment_states.running = running or random_state_count()
    experiment_states.cleaning = cleaning or random_state_count()
    experiment_states.aborted = aborted or random_state_count()
    experiment_states.crashed = crashed or random_state_count()
    experiment_states.failed = failed or random_state_count()
    experiment_states.succeeded = succeeded or random_state_count()
    experiment_states.notResponding = not_responding or random_state_count()
    return experiment_states


def a_property():
    p = MagicMock()
    p.key = a_string()
    p.value = a_string()
    return p


def a_parameter():
    p = MagicMock()
    p.id = a_uuid_string()
    p.name = a_string()
    p.parameterType = 'double'
    p.value = str(uniform(-100, 100))
    return p


def a_channel_value():
    cv = MagicMock()
    cv.channelId = a_uuid_string()
    cv.channelName = a_string()
    cv.channelType = 'numeric'
    cv.x = uniform(1, 100)
    cv.y = str(uniform(1, 100))
    return cv


def an_experiment_leaderboard_entry_dto():
    entry = MagicMock()
    entry.entryType = 'experiment'
    entry.id = a_uuid_string()
    entry.shortId = a_string()
    entry.projectId = a_uuid_string()
    entry.state = 'succeeded'
    entry.experimentStates = an_experiment_states(succeeded=1)
    entry.responding = True
    entry.name = a_string()
    entry.organizationName = a_string()
    entry.projectName = a_string()
    entry.description = a_string()
    entry.timeOfCreation = a_timestamp()
    entry.timeOfCompletion = a_timestamp()
    entry.runningTime = randint(1, 1000)
    entry.owner = a_string()
    entry.size = randint(1, 1000)
    entry.tags = [a_string(), a_string()]
    entry.environment = a_string()
    entry.workerType = a_string()
    entry.hostname = a_string()
    entry.sourceSize = randint(1, 1000)
    entry.sourceMd5 = a_string()
    entry.commitId = a_string()
    entry.properties = [a_property(), a_property()]
    entry.parameters = [a_parameter(), a_parameter()]
    entry.channelsLastValues = [a_channel_value(), a_channel_value()]
    entry.trashed = False
    entry.deleted = False
    entry.isBestExperiment = False
    return entry



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging
import uuid

from typing import TYPE_CHECKING, Optional, List

from neptune.new.attributes.constants import SYSTEM_FAILED_ATTRIBUTE_PATH
from neptune.new.internal.background_job import BackgroundJob
from neptune.new.internal.utils.uncaught_exception_handler import instance as traceback_handler

if TYPE_CHECKING:
    from neptune.new.run import Run

_logger = logging.getLogger(__name__)


class TracebackJob(BackgroundJob):

    def __init__(self, path: str, fail_on_exception: bool = True):
        self._uuid = uuid.uuid4()
        self._started = False
        self._path = path
        self._fail_on_exception = fail_on_exception

    def start(self, run: 'Run'):
        if not self._started:
            path = self._path
            fail_on_exception = self._fail_on_exception

            def log_traceback(stacktrace_lines: List[str]):
                run[path].log(stacktrace_lines)
                if fail_on_exception:
                    run[SYSTEM_FAILED_ATTRIBUTE_PATH] = True

            traceback_handler.register(self._uuid, log_traceback)
        self._started = True

    def stop(self):
        traceback_handler.unregister(self._uuid)

    def join(self, seconds: Optional[float] = None):
        pass



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unittest
from datetime import datetime

import mock

from neptune.internal.streams.channel_writer import ChannelWriter


class TestChannelWriter(unittest.TestCase):

    def test_write_data_to_channel_writer(self):
        # given
        experiment = mock.MagicMock()
        experiment.get_system_properties.return_value = {"created": datetime.now()}
        channel_name = 'a channel name'
        writer = ChannelWriter(experiment, channel_name)

        # when
        writer.write('some\ndata')

        # then
        # pylint: disable=protected-access
        experiment._channels_values_sender.send.assert_called_once()

    @mock.patch('neptune.internal.streams.channel_writer.datetime')
    def test_write_data_with_low_resolution_datetime_now(self, dt):
        # given
        experiment = mock.MagicMock()
        experiment.get_system_properties.return_value = {"created": datetime(2022, 2, 2, 2, 2, 2, 2)}
        channel_name = 'a channel name'
        writer = ChannelWriter(experiment, channel_name)

        # and
        dt.now.return_value = datetime(2022, 2, 2, 2, 2, 2, 3)

        # when
        writer.write('text1\ntext2\n')

        # then
        # pylint: disable=protected-access
        x_to_text = self._extract_x_to_text_from_calls(experiment._channels_values_sender.send.call_args_list)
        self.assertEqual(x_to_text, {0.001: 'text1', 0.002: 'text2'})

    @mock.patch('neptune.internal.streams.channel_writer.datetime')
    def test_write_data_with_high_resolution_datetime_now(self, dt):
        # given
        experiment = mock.MagicMock()
        experiment.get_system_properties.return_value = {"created": datetime(2022, 2, 2, 2, 2, 2, 2)}
        channel_name = 'a channel name'
        writer = ChannelWriter(experiment, channel_name)

        # when
        dt.now.return_value = datetime(2022, 2, 2, 2, 2, 2, 4)
        writer.write('text1\n')
        dt.now.return_value = datetime(2022, 2, 2, 2, 2, 2, 5)
        writer.write('text2\n')

        # then
        # pylint: disable=protected-access
        x_to_text = self._extract_x_to_text_from_calls(experiment._channels_values_sender.send.call_args_list)
        self.assertEqual(x_to_text, {0.002: 'text1', 0.003: 'text2'})

    @staticmethod
    def _extract_x_to_text_from_calls(calls):
        channel_values = [kwargs['channel_value'] for (_, kwargs) in calls]
        return dict((v.x, v.y['text_value']) for v in channel_values)


if __name__ == '__main__':
    unittest.main()



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
from io import IOBase
from typing import TypeVar, TYPE_CHECKING, Optional, Union

from neptune.new.internal.utils.images import get_image_content, get_html_content, get_pickle_content, is_pil_image, \
    is_matplotlib_figure, is_plotly_figure, is_altair_chart, is_bokeh_figure, is_numpy_array, is_pandas_dataframe


from neptune.new.internal.utils import verify_type, get_stream_content
from neptune.new.types.atoms.atom import Atom

if TYPE_CHECKING:
    from neptune.new.types.value_visitor import ValueVisitor

Ret = TypeVar('Ret')


class File(Atom):

    def __init__(self,
                 path: Optional[str] = None,
                 content: Optional[bytes] = None,
                 extension: Optional[str] = None):
        verify_type("path", path, (str, type(None)))
        verify_type("content", content, (bytes, type(None)))
        verify_type("extension", extension, (str, type(None)))

        if path is not None and content is not None:
            raise ValueError("path and content are mutually exclusive")
        if path is None and content is None:
            raise ValueError("path or content is required")

        self.path = path
        self.content = content

        if extension is None and path is not None:
            try:
                ext = os.path.splitext(path)[1]
                self.extension = ext[1:] if ext else ""
            except ValueError:
                self.extension = ""
        else:
            self.extension = extension or ""

    def accept(self, visitor: 'ValueVisitor[Ret]') -> Ret:
        return visitor.visit_file(self)

    def __str__(self):
        if self.path is not None:
            return "File(path={})".format(str(self.path))
        else:
            return "File(content=...)"

    @staticmethod
    def from_content(content: Union[str, bytes], extension: Optional[str] = None) -> 'File':
        """Factory method for creating File value objects directly from binary and text content.

        In the case of text content, UTF-8 encoding will be used.

        Args:
            content	(str or bytes): Text or binary content to stored in the `File` value object.
            extension (str, optional, default is None): Extension of the created file.
                File will be used for interpreting the type of content for visualization.
                If `None` it will be bin for binary content and txt for text content.
                Defaults to `None`.

        Returns:
            ``File``: value object created from the content

        You may also want to check `from_content docs page`_.

        .. _from_content docs page:
           https://docs.neptune.ai/api-reference/field-types#from_content
        """
        if isinstance(content, str):
            ext = "txt"
            content = content.encode("utf-8")
        else:
            ext = "bin"
        return File(content=content, extension=extension or ext)

    @staticmethod
    def from_stream(stream: IOBase, seek: Optional[int] = 0, extension: Optional[str] = None) -> 'File':
        """Factory method for creating File value objects directly from binary and text streams.

        In the case of text stream, UTF-8 encoding will be used.

        Args:
            stream (IOBase): Stream to be converted.
            seek (int, optional): See IOBase documentation.
                Defaults to `0`.
            extension (str, optional): Extension of the file created that will be used for interpreting the type
                of content for visualization.
                If `None` it will be bin for binary stream and txt for text stream.
                Defaults to `None`.

        Returns:
            ``File``: value object created from the stream.

        You may also want to check `from_stream docs page`_ and `IOBase documentation`_.

        .. _from_stream docs page:
           https://docs.neptune.ai/api-reference/field-types#from_stream
        .. _IOBase documentation:
            https://docs.python.org/3/library/io.html#io.IOBase
        """
        verify_type("stream", stream, IOBase)
        content, stream_default_ext = get_stream_content(stream, seek)
        return File(content=content, extension=extension or stream_default_ext)

    @staticmethod
    def as_image(image) -> 'File':
        """Static method for converting image objects or image-like objects to an image File value object.

        This way you can upload `Matplotlib` figures, `PIL` images, `NumPy` arrays, as static images.

        Args:
            image: Image-like object to be converted.
                Supported are `PyTorch` tensors, `TensorFlow/Keras` tensors, `NumPy` arrays, `PIL` images
                and `Matplotlib` figures.

        Returns:
            ``File``: value object with converted image

        Examples:
            >>> import neptune.new as neptune
            >>> from neptune.new.types import File
            >>> run = neptune.init()

            Convert NumPy array to File value object and upload it

            >>> run["train/prediction_example"].upload(File.as_image(numpy_array))

            Convert PIL image to File value object and upload it

            >>> pil_file = File.as_image(pil_image)
            >>> run["dataset/data_sample/img1"].upload(pil_file)

            You can upload PIL image without explicit conversion

            >>> run["dataset/data_sample/img2"].upload(pil_image)

        You may also want to check `as_image docs page`_.

        .. _as_image docs page:
           https://docs.neptune.ai/api-reference/field-types#as_image
        """
        content_bytes = get_image_content(image)
        return File.from_content(content_bytes if content_bytes is not None else b"", extension="png")

    @staticmethod
    def as_html(chart) -> 'File':
        """Converts an object to an HTML File value object.

        This way you can upload `Altair`, `Bokeh`, `Plotly`, `Matplotlib` interactive charts
        or upload directly `Pandas` `DataFrame` objects to explore them in Neptune UI.

        Args:
            chart: An object to be converted.
                Supported are `Altair`, `Bokeh`, `Plotly`, `Matplotlib` interactive charts,
                and `Pandas` `DataFrame` objects.

        Returns:
            ``File``: value object with converted object.

        Examples:
            >>> import neptune.new as neptune
            >>> from neptune.new.types import File
            >>> run = neptune.init()

            Convert Pandas DataFrame to File value object and upload it

            >>> run["train/results"].upload(File.as_html(df_predictions))

            Convert Altair interactive chart to File value object and upload it

            >>> altair_file = File.as_html(altair_chart)
            >>> run["dataset/data_sample/img1"].upload(altair_file)

            You can upload Altair interactive chart without explicit conversion

            >>> run["dataset/data_sample/img2"].upload(altair_chart)

        You may also want to check `as_html docs page`_.

        .. _as_html docs page:
           https://docs.neptune.ai/api-reference/field-types#as_html
        """
        content = get_html_content(chart)
        return File.from_content(content if content is not None else "", extension="html")

    @staticmethod
    def as_pickle(obj) -> 'File':
        """Pickles a Python object and stores it in `File` value object.

        This way you can upload any Python object for future use.

        Args:
            obj: An object to be converted.
                Supported are `Altair`, `Bokeh`, `Plotly`, `Matplotlib` interactive charts,
                and `Pandas` `DataFrame` objects.

        Returns:
            ``File``: value object with pickled object.

        Examples:
            >>> import neptune.new as neptune
            >>> from neptune.new.types import File
            >>> run = neptune.init()

            Pickle model object and upload it

            >>> run["results/pickled_model"].upload(File.as_pickle(trained_model))

        You may also want to check `as_pickle docs page`_.

        .. _as_pickle docs page:
           https://docs.neptune.ai/api-reference/field-types#as_pickle
        """
        content = get_pickle_content(obj)
        return File.from_content(content if content is not None else b"", extension="pkl")

    @staticmethod
    def create_from(value) -> 'File':
        if isinstance(value, str):
            return File(path=value)
        elif is_pil_image(value) or is_matplotlib_figure(value):
            return File.as_image(value)
        elif is_plotly_figure(value) or is_altair_chart(value) or is_bokeh_figure(value):
            return File.as_html(value)
        elif is_numpy_array(value):
            raise TypeError("Value of type {} is not supported. Please use File.as_image().".format(type(value)))
        elif is_pandas_dataframe(value):
            raise TypeError("Value of type {} is not supported. Please use File.as_html().".format(type(value)))
        elif isinstance(value, File):
            return value
        raise TypeError("Value of type {} is not supported.".format(type(value)))

    @staticmethod
    def is_convertable(value):
        return is_pil_image(value) \
               or is_matplotlib_figure(value) \
               or is_plotly_figure(value) \
               or is_altair_chart(value) \
               or is_bokeh_figure(value) \
               or is_numpy_array(value) \
               or is_pandas_dataframe(value) \
               or isinstance(value, File)



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# pylint: disable=protected-access
import pickle
from io import StringIO, BytesIO

import numpy
from PIL import Image

from neptune.new.internal.utils.images import _get_pil_image_data
from neptune.new.types import File

from tests.neptune.new.attributes.test_attribute_base import TestAttributeBase


class TestFile(TestAttributeBase):

    def test_create_from_path(self):
        file = File("some/path.ext")
        self.assertEqual("some/path.ext", file.path)
        self.assertEqual(None, file.content)
        self.assertEqual("ext", file.extension)

        file = File("some/path.txt.ext")
        self.assertEqual("some/path.txt.ext", file.path)
        self.assertEqual(None, file.content)
        self.assertEqual("ext", file.extension)

        file = File("so.me/path")
        self.assertEqual("so.me/path", file.path)
        self.assertEqual(None, file.content)
        self.assertEqual("", file.extension)

        file = File("some/path.ext", extension="txt")
        self.assertEqual("some/path.ext", file.path)
        self.assertEqual(None, file.content)
        self.assertEqual("txt", file.extension)

    def test_create_from_string_content(self):
        file = File.from_content("some_content")
        self.assertEqual(None, file.path)
        self.assertEqual("some_content".encode("utf-8"), file.content)
        self.assertEqual("txt", file.extension)

        file = File.from_content("some_content", extension="png")
        self.assertEqual(None, file.path)
        self.assertEqual("some_content".encode("utf-8"), file.content)
        self.assertEqual("png", file.extension)

    def test_create_from_bytes_content(self):
        file = File.from_content(b"some_content")
        self.assertEqual(None, file.path)
        self.assertEqual(b"some_content", file.content)
        self.assertEqual("bin", file.extension)

        file = File.from_content(b"some_content", extension="png")
        self.assertEqual(None, file.path)
        self.assertEqual(b"some_content", file.content)
        self.assertEqual("png", file.extension)

    def test_create_from_string_io(self):
        file = File.from_stream(StringIO("aaabbbccc"))
        self.assertEqual(None, file.path)
        self.assertEqual(b"aaabbbccc", file.content)
        self.assertEqual("txt", file.extension)

        stream = StringIO("aaabbbccc")
        stream.seek(3)
        file = File.from_stream(stream)
        self.assertEqual(None, file.path)
        self.assertEqual(b"aaabbbccc", file.content)
        self.assertEqual("txt", file.extension)

        file = File.from_stream(StringIO("aaabbbccc"), extension="png")
        self.assertEqual(None, file.path)
        self.assertEqual(b"aaabbbccc", file.content)
        self.assertEqual("png", file.extension)

        file = File.from_stream(StringIO("aaabbbccc"), seek=5)
        self.assertEqual(None, file.path)
        self.assertEqual(b"bccc", file.content)
        self.assertEqual("txt", file.extension)

    def test_create_from_bytes_io(self):
        file = File.from_stream(BytesIO(b"aaabbbccc"))
        self.assertEqual(None, file.path)
        self.assertEqual(b"aaabbbccc", file.content)
        self.assertEqual("bin", file.extension)

        stream = BytesIO(b"aaabbbccc")
        stream.seek(3)
        file = File.from_stream(stream)
        self.assertEqual(None, file.path)
        self.assertEqual(b"aaabbbccc", file.content)
        self.assertEqual("bin", file.extension)

        file = File.from_stream(BytesIO(b"aaabbbccc"), extension="png")
        self.assertEqual(None, file.path)
        self.assertEqual(b"aaabbbccc", file.content)
        self.assertEqual("png", file.extension)

        file = File.from_stream(BytesIO(b"aaabbbccc"), seek=5)
        self.assertEqual(None, file.path)
        self.assertEqual(b"bccc", file.content)
        self.assertEqual("bin", file.extension)

    def test_as_image(self):
        # given
        image_array = numpy.random.rand(10, 10) * 255
        expected_image = Image.fromarray(image_array.astype(numpy.uint8))

        # when
        file = File.as_image(expected_image)

        # then
        self.assertEqual(file.extension, "png")
        self.assertEqual(file.content, _get_pil_image_data(expected_image))

    def test_as_html(self):
        # given
        from bokeh.plotting import figure
        # given
        p = figure(plot_width=400, plot_height=400)
        p.circle(size=20, color="navy", alpha=0.5)

        # when
        file = File.as_html(p)

        # then
        self.assertEqual(file.extension, "html")
        self.assertTrue(file.content.startswith('\n\n\n\n<!DOCTYPE html>\n<html lang="en">'.encode("utf-8")))

    def test_as_pickle(self):
        # given
        obj = {"a": [b"xyz", 34], "b": 1246}

        # when
        file = File.as_pickle(obj)

        # then
        self.assertEqual(file.extension, "pkl")
        self.assertEqual(file.content, pickle.dumps(obj))

    def test_raise_exception_in_constructor(self):
        with self.assertRaises(ValueError):
            File(path="path", content=b"some_content")
        with self.assertRaises(ValueError):
            File()



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Union

from neptune.new.internal.operation import AssignBool
from neptune.new.types.atoms.boolean import Boolean as BooleanVal
from neptune.new.attributes.atoms.atom import Atom


class Boolean(Atom):

    def assign(self, value: Union[BooleanVal, bool], wait: bool = False):
        if not isinstance(value, BooleanVal):
            value = BooleanVal(value)

        with self._run.lock():
            self._enqueue_operation(AssignBool(self._path, value.value), wait)

    def fetch(self) -> float:
        # pylint: disable=protected-access
        val = self._backend.get_bool_attribute(self._run_uuid, self._path)
        return val.value



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import uuid
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import Optional, List, Set, Any

from packaging.version import Version


class Project:

    def __init__(self, _uuid: uuid.UUID, name: str, workspace: str):
        self.uuid = _uuid
        self.name = name
        self.workspace = workspace


class Workspace:

    def __init__(self, _uuid: uuid.UUID, name: str):
        self.uuid = _uuid
        self.name = name


@dataclass
class ApiRun:
    uuid: uuid.UUID
    short_id: str
    workspace: str
    project_name: str
    trashed: bool


class ClientConfig(object):

    def __init__(self,
                 api_url: str,
                 display_url: str,
                 min_recommended_version: Optional[Version],
                 min_compatible_version: Optional[Version],
                 max_compatible_version: Optional[Version]):
        self._api_url = api_url
        self._display_url = display_url
        self._min_recommended_version = min_recommended_version
        self._min_compatible_version = min_compatible_version
        self._max_compatible_version = max_compatible_version

    @property
    def api_url(self) -> str:
        return self._api_url

    @property
    def display_url(self) -> str:
        return self._display_url

    @property
    def min_recommended_version(self) -> Optional[Version]:
        return self._min_recommended_version

    @property
    def min_compatible_version(self) -> Optional[Version]:
        return self._min_compatible_version

    @property
    def max_compatible_version(self) -> Optional[Version]:
        return self._max_compatible_version


class AttributeType(Enum):
    FLOAT = "float"
    INT = "int"
    BOOL = "bool"
    STRING = "string"
    DATETIME = "datetime"
    FILE = "file"
    FILE_SET = "fileSet"
    FLOAT_SERIES = "floatSeries"
    STRING_SERIES = "stringSeries"
    IMAGE_SERIES = "imageSeries"
    STRING_SET = "stringSet"
    GIT_REF = "gitRef"
    RUN_STATE = "experimentState"
    NOTEBOOK_REF = "notebookRef"


@dataclass
class Attribute:
    path: str
    type: AttributeType


@dataclass
class AttributeWithProperties:
    path: str
    type: AttributeType
    properties: Any


@dataclass
class LeaderboardEntry:
    id: uuid.UUID
    attributes: List[AttributeWithProperties]


@dataclass
class StringPointValue:
    timestampMillis: int
    step: float
    value: str


@dataclass
class ImageSeriesValues:
    totalItemCount: int


@dataclass
class StringSeriesValues:
    totalItemCount: int
    values: List[StringPointValue]


@dataclass
class FloatPointValue:
    timestampMillis: int
    step: float
    value: float


@dataclass
class FloatSeriesValues:
    totalItemCount: int
    values: List[FloatPointValue]


@dataclass
class FloatAttribute:
    value: float


@dataclass
class IntAttribute:
    value: int


@dataclass
class BoolAttribute:
    value: bool


@dataclass
class FileAttribute:
    name: str
    ext: str
    size: int


@dataclass
class StringAttribute:
    value: str


@dataclass
class DatetimeAttribute:
    value: datetime


@dataclass
class FloatSeriesAttribute:
    last: Optional[float]


@dataclass
class StringSeriesAttribute:
    last: Optional[str]


@dataclass
class StringSetAttribute:
    values: Set[str]



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unittest
from tempfile import NamedTemporaryFile

from neptune.new.internal.utils.json_file_splitter import JsonFileSplitter


class TestJsonFileSplitter(unittest.TestCase):

    def test_simple_file(self):
        content = """
{
    "a": 5,
    "b": "text"
}
{
    "a": 13
}
{}
""".lstrip()

        with self._create_file(content) as file:
            splitter = JsonFileSplitter(file.name)
            self.assertEqual(splitter.get(), {"a": 5, "b": "text"})
            self.assertEqual(splitter.get(), {"a": 13})
            self.assertEqual(splitter.get(), {})
            self.assertEqual(splitter.get(), None)
            splitter.close()

    def test_append(self):
        content1 = """
            {
                "a": 5,
                "b": "text"
            }
            {
                "a": 13
            }"""

        content2 = """
            {
                "q": 555,
                "r": "something"
            }
            {
                "a": {
                    "b": [1, 2, 3]
                }
            }
            {}"""

        with self._create_file(content1) as file:
            splitter = JsonFileSplitter(file.name)
            self.assertEqual(splitter.get(), {"a": 5, "b": "text"})
            self.assertEqual(splitter.get(), {"a": 13})
            self.assertEqual(splitter.get(), None)
            file.write(content2)
            file.flush()
            self.assertEqual(splitter.get(), {"q": 555, "r": "something"})
            self.assertEqual(splitter.get(), {"a": {"b": [1, 2, 3]}})
            self.assertEqual(splitter.get(), {})
            self.assertEqual(splitter.get(), None)
            splitter.close()

    def test_append_cut_json(self):
        content1 = """
            {
                "a": 5,
                "b": "text"
            }
            {
                "a": 1"""

        content2 = """55,
                "r": "something"
            }
            {
                "a": {
                    "b": [1, 2, 3]
                }
            }"""

        with self._create_file(content1) as file:
            splitter = JsonFileSplitter(file.name)
            self.assertEqual(splitter.get(), {"a": 5, "b": "text"})
            self.assertEqual(splitter.get(), None)
            file.write(content2)
            file.flush()
            self.assertEqual(splitter.get(), {"a": 155, "r": "something"})
            self.assertEqual(splitter.get(), {"a": {"b": [1, 2, 3]}})
            self.assertEqual(splitter.get(), None)
            splitter.close()

    def test_big_json(self):
        content = """
{
    "a": 5,
    "b": "text"
}
{
    "a": "%s",
    "b": "%s"
}
{}
""".lstrip() % ("x" * JsonFileSplitter.BUFFER_SIZE * 2, "y" * JsonFileSplitter.BUFFER_SIZE * 2)

        with self._create_file(content) as file:
            splitter = JsonFileSplitter(file.name)
            self.assertEqual(splitter.get(), {"a": 5, "b": "text"})
            self.assertEqual(splitter.get(), {"a": "x" * JsonFileSplitter.BUFFER_SIZE * 2,
                                              "b": "y" * JsonFileSplitter.BUFFER_SIZE * 2})
            self.assertEqual(splitter.get(), {})
            self.assertEqual(splitter.get(), None)
            splitter.close()

    @staticmethod
    def _create_file(content):
        file = NamedTemporaryFile("w")
        file.write(content)
        file.flush()
        return file



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# pylint: disable=protected-access
from typing import Optional, Iterable

from mock import MagicMock, call, patch

from neptune.new.internal.operation import ClearFloatLog, LogFloats, Operation, ConfigFloatSeries, ClearStringLog
from neptune.new.attributes.series.float_series import FloatSeries, FloatSeriesVal
from neptune.new.attributes.series.string_series import StringSeries, StringSeriesVal
from neptune.new.attributes.series.series import Series

from tests.neptune.new.attributes.test_attribute_base import TestAttributeBase


@patch("time.time", new=TestAttributeBase._now)
class TestSeries(TestAttributeBase):

    def test_assign(self):
        value = FloatSeriesVal([17, 3.6], min=0, max=100, unit="%")
        expected = [LogFloats.ValueType(17, None, self._now()), LogFloats.ValueType(3.6, None, self._now())]

        processor = MagicMock()
        exp, path, wait = self._create_run(processor), self._random_path(), self._random_wait()
        var = FloatSeries(exp, path)
        var.assign(value, wait=wait)
        self.assertEqual(3, processor.enqueue_operation.call_count)
        processor.enqueue_operation.assert_has_calls([
            call(ConfigFloatSeries(path, min=0, max=100, unit="%"), False),
            call(ClearFloatLog(path), False),
            call(LogFloats(path, expected), wait)
        ])

    def test_assign_empty(self):
        processor = MagicMock()
        exp, path, wait = self._create_run(processor), self._random_path(), self._random_wait()
        var = StringSeries(exp, path)
        var.assign(StringSeriesVal([]), wait=wait)
        processor.enqueue_operation.assert_called_once_with(ClearStringLog(path), wait)

    def test_log(self):
        value_and_expected = [
            (13, [LogFloats.ValueType(13, None, self._now())]),
            (15.3, [LogFloats.ValueType(15.3, None, self._now())]),
            ([], []),
            ([1, 9, 7], [LogFloats.ValueType(1, None, self._now()),
                         LogFloats.ValueType(9, None, self._now()),
                         LogFloats.ValueType(7, None, self._now())]),
            ((1, 9, 7), [LogFloats.ValueType(1, None, self._now()),
                         LogFloats.ValueType(9, None, self._now()),
                         LogFloats.ValueType(7, None, self._now())]),
            ({1, 9, 7}, [LogFloats.ValueType(1, None, self._now()),
                         LogFloats.ValueType(9, None, self._now()),
                         LogFloats.ValueType(7, None, self._now())])
        ]

        for value, expected in value_and_expected:
            processor = MagicMock()
            exp, path, wait = self._create_run(processor), self._random_path(), self._random_wait()
            var = FloatSeries(exp, path)
            var.log(value, wait=wait)
            processor.enqueue_operation.assert_called_once_with(LogFloats(path, expected), wait)

    def test_log_with_step(self):
        value_step_and_expected = [
            (13, 5.3, LogFloats.ValueType(13, 5.3, self._now())),
            (15.3, 10, LogFloats.ValueType(15.3, 10, self._now())),
            ([13], 5.3, LogFloats.ValueType(13, 5.3, self._now())),
            ((13,), 5.3, LogFloats.ValueType(13, 5.3, self._now())),
            ({13}, 5.3, LogFloats.ValueType(13, 5.3, self._now()))
        ]

        for value, step, expected in value_step_and_expected:
            processor = MagicMock()
            exp, path, wait = self._create_run(processor), self._random_path(), self._random_wait()
            var = FloatSeries(exp, path)
            var.log(value, step=step, wait=wait)
            processor.enqueue_operation.assert_called_once_with(LogFloats(path, [expected]), wait)

    def test_log_with_timestamp(self):
        value_step_and_expected = [
            (13, 5.3, LogFloats.ValueType(13, None, 5.3)),
            (15.3, 10, LogFloats.ValueType(15.3, None, 10))
        ]

        for value, ts, expected in value_step_and_expected:
            processor = MagicMock()
            exp, path, wait = self._create_run(processor), self._random_path(), self._random_wait()
            var = FloatSeries(exp, path)
            var.log(value, timestamp=ts, wait=wait)
            processor.enqueue_operation.assert_called_once_with(LogFloats(path, [expected]), wait)

    def test_log_value_errors(self):
        processor = MagicMock()
        exp, path = self._create_run(processor), self._random_path()
        attr = FloatSeries(exp, path)

        with self.assertRaises(ValueError):
            attr.log(["str", 5])
        with self.assertRaises(ValueError):
            attr.log([5, 10], step=10)
        with self.assertRaises(TypeError):
            attr.log(5, step="str")
        with self.assertRaises(TypeError):
            attr.log(5, timestamp="str")

    def test_clear(self):
        processor = MagicMock()
        exp, path, wait = self._create_run(processor), self._random_path(), self._random_wait()
        var = FloatSeries(exp, path)
        var.clear(wait=wait)
        processor.enqueue_operation.assert_called_once_with(ClearFloatLog(path), wait)

    class SeriesTestClass(Series[FloatSeriesVal, int]):

        def _get_log_operation_from_value(self,
                                          value: FloatSeriesVal,
                                          step: Optional[float],
                                          timestamp: float) -> Operation:
            values = [LogFloats.ValueType(val, step=step, ts=timestamp) for val in value.values]
            return LogFloats(self._path, values)

        def _get_clear_operation(self) -> Operation:
            return ClearFloatLog(self._path)

        # pylint: disable=unused-argument
        def _data_to_value(self, values: Iterable, **kwargs) -> FloatSeriesVal:
            return FloatSeriesVal(values)

        def _is_value_type(self, value) -> bool:
            return isinstance(value, FloatSeriesVal)



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unittest

from neptune.new.internal.utils import verify_type, verify_collection_type


class TestUtils(unittest.TestCase):

    def test_verify_type(self):
        verify_type("arg", "string", str)

    def test_verify_type_failed(self):
        with self.assertRaises(TypeError):
            verify_type("arg", 5, str)

    def test_verify_type_tuple(self):
        verify_type("arg", "string", (int, float, str))

    def test_verify_type_tuple_failed(self):
        with self.assertRaises(TypeError):
            verify_type("arg", 5, (str, type(None), float))

    def test_verify_collection_type(self):
        verify_collection_type("arg", ["string", "aaa", 5, 1, "q"], (int, str))

    def test_verify_collection_type_failed(self):
        with self.assertRaises(TypeError):
            verify_collection_type("arg", "string", (int, str))

    def test_verify_collection_type_failed_element(self):
        with self.assertRaises(TypeError):
            verify_collection_type("arg", ["string", 3, "a", 4., 1], (int, str))



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from mock import MagicMock

from neptune.internal.hardware.gauges.cpu import SystemCpuUsageGauge
from neptune.internal.hardware.gauges.gauge_factory import GaugeFactory
from neptune.internal.hardware.gauges.gpu import GpuMemoryGauge, GpuUsageGauge
from neptune.internal.hardware.gauges.memory import SystemMemoryUsageGauge


class GaugesFixture(object):
    def __init__(self):
        self.gauge_factory = MagicMock(spec_set=GaugeFactory)

        self.cpu_gauge_value = 1.0
        self.memory_gauge_value = 2.0

        self.gpu0_usage_gauge_value = 3.0
        self.gpu1_usage_gauge_value = 4.0

        self.gpu0_memory_gauge_value = 5.0
        self.gpu1_memory_gauge_value = 6.0

        cpu_gauge = MagicMock(wraps=SystemCpuUsageGauge())
        cpu_gauge.value.return_value = self.cpu_gauge_value
        self.gauge_factory.create_cpu_usage_gauge.return_value = cpu_gauge

        ram_gauge = MagicMock(wraps=SystemMemoryUsageGauge())
        ram_gauge.value.return_value = self.memory_gauge_value
        self.gauge_factory.create_memory_usage_gauge.return_value = ram_gauge

        gpu_usage_gauge_0 = MagicMock(wraps=GpuUsageGauge(card_index=0))
        gpu_usage_gauge_0.value.return_value = self.gpu0_usage_gauge_value

        gpu_usage_gauge_2 = MagicMock(wraps=GpuUsageGauge(card_index=2))
        gpu_usage_gauge_2.value.return_value = self.gpu1_usage_gauge_value

        self.gauge_factory.create_gpu_usage_gauge.side_effect = \
            lambda card_index: gpu_usage_gauge_0 if card_index == 0 else gpu_usage_gauge_2

        gpu_memory_gauge_0 = MagicMock(wraps=GpuMemoryGauge(card_index=0))
        gpu_memory_gauge_0.value.return_value = self.gpu0_memory_gauge_value

        gpu_memory_gauge_2 = MagicMock(wraps=GpuMemoryGauge(card_index=2))
        gpu_memory_gauge_2.value.return_value = self.gpu1_memory_gauge_value
        self.gauge_factory.create_gpu_memory_gauge.side_effect = \
            lambda card_index: gpu_memory_gauge_0 if card_index == 0 else gpu_memory_gauge_2



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Optional, Iterable, List, TYPE_CHECKING

import click

from neptune.new.attributes.series.fetchable_series import FetchableSeries
from neptune.new.internal.backends.api_model import StringSeriesValues
from neptune.new.internal.utils.paths import path_to_str
from neptune.new.types.series.string_series import StringSeries as StringSeriesVal

from neptune.new.internal.operation import LogStrings, ClearStringLog, Operation
from neptune.new.attributes.series.series import Series

if TYPE_CHECKING:
    from neptune.new.run import Run

Val = StringSeriesVal
Data = str

MAX_STRING_SERIES_VALUE_LENGTH = 1000


class StringSeries(Series[Val, Data], FetchableSeries[StringSeriesValues]):

    def __init__(self, run: 'Run', path: List[str]):
        super().__init__(run, path)
        self._value_truncation_occurred = False

    def _get_log_operation_from_value(self, value: Val, step: Optional[float], timestamp: float) -> Operation:
        values = [v[:MAX_STRING_SERIES_VALUE_LENGTH] for v in value.values]
        if not self._value_truncation_occurred \
                and any([len(v) > MAX_STRING_SERIES_VALUE_LENGTH for v in value.values]):
            # the first truncation
            self._value_truncation_occurred = True
            click.echo(f"Warning: string series '{ path_to_str(self._path)}' value was "
                       f"longer than {MAX_STRING_SERIES_VALUE_LENGTH} characters and was truncated. "
                       f"This warning is printed only once per series.", err=True)

        values = [LogStrings.ValueType(val, step=step, ts=timestamp) for val in values]
        return LogStrings(self._path, values)

    def _get_clear_operation(self) -> Operation:
        return ClearStringLog(self._path)

    def _data_to_value(self, values: Iterable, **kwargs) -> Val:
        if kwargs:
            click.echo("Warning: unexpected arguments ({kwargs}) in StringSeries".format(kwargs=kwargs), err=True)
        return StringSeriesVal(values)

    def _is_value_type(self, value) -> bool:
        return isinstance(value, StringSeriesVal)

    def fetch_last(self) -> str:
        # pylint: disable=protected-access
        val = self._backend.get_string_series_attribute(self._run_uuid, self._path)
        return val.last

    def _fetch_values_from_backend(self, offset, limit) -> StringSeriesValues:
        return self._backend.get_string_series_values(self._run_uuid, self._path, offset, limit)



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import base64
import io
import logging
import pickle
import warnings
from io import StringIO, BytesIO
from typing import Optional

import click
from packaging import version
from pandas import DataFrame

from neptune.new.exceptions import PlotlyIncompatibilityException
from neptune.new.internal.utils import limits

_logger = logging.getLogger(__name__)

try:
    from numpy import ndarray as numpy_ndarray, array as numpy_array, uint8 as numpy_uint8
except ImportError:
    numpy_ndarray = None
    numpy_array = None
    numpy_uint8 = None

try:
    from PIL.Image import Image as PILImage, fromarray as pilimage_fromarray
except ImportError:
    PILImage = None

    def pilimage_fromarray():
        pass


def get_image_content(image) -> Optional[bytes]:
    content = _image_to_bytes(image)

    if limits.image_size_exceeds_limit(len(content)):
        return None

    return content


def get_html_content(chart) -> Optional[str]:
    content = _to_html(chart)

    if limits.file_size_exceeds_limit(len(content)):
        return None

    return content


def get_pickle_content(obj) -> Optional[bytes]:
    content = _export_pickle(obj)

    if limits.file_size_exceeds_limit(len(content)):
        return None

    return content


def _image_to_bytes(image) -> bytes:
    if image is None:
        raise ValueError("image is None")

    elif is_numpy_array(image):
        return _get_numpy_as_image(image)

    elif is_pil_image(image):
        return _get_pil_image_data(image)

    elif is_matplotlib_figure(image):
        return _get_figure_image_data(image)

    elif _is_torch_tensor(image):
        return _get_numpy_as_image(image.detach().numpy())

    elif _is_tensorflow_tensor(image):
        return _get_numpy_as_image(image.numpy())

    raise TypeError("image is {}".format(type(image)))


def _to_html(chart) -> str:
    if _is_matplotlib_pyplot(chart):
        chart = chart.gcf()

    if is_matplotlib_figure(chart):
        try:
            chart = _matplotlib_to_plotly(chart)
            return _export_plotly_figure(chart)
        except ImportError:
            print("Plotly not installed. Logging plot as an image.")
            return _image_content_to_html(_get_figure_image_data(chart))
        except UserWarning:
            print("Couldn't convert Matplotlib plot to interactive Plotly plot. Logging plot as an image instead.")
            return _image_content_to_html(_get_figure_image_data(chart))

    elif is_pandas_dataframe(chart):
        return _export_pandas_dataframe_to_html(chart)

    elif is_plotly_figure(chart):
        return _export_plotly_figure(chart)

    elif is_altair_chart(chart):
        return _export_altair_chart(chart)

    elif is_bokeh_figure(chart):
        return _export_bokeh_figure(chart)

    else:
        raise ValueError("Currently supported are matplotlib, plotly, altair, and bokeh figures")


def _matplotlib_to_plotly(chart):
    # pylint: disable=import-outside-toplevel
    import plotly
    import matplotlib

    # When Plotly cannot accurately convert a matplotlib plot, it emits a warning.
    # Then we want to fallback on logging the plot as an image.
    #
    # E.g. when trying to convert a Seaborn confusion matrix or a hist2d, it emits a UserWarning with message
    # "Dang! That path collection is out of this world. I totally don't know what to do with it yet!
    # Plotly can only import path collections linked to 'data' coordinates"
    plotly_version = plotly.__version__
    matplotlib_version = matplotlib.__version__
    if version.parse(matplotlib_version) >= version.parse("3.3.0"):
        raise PlotlyIncompatibilityException(matplotlib_version, plotly_version)

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "error",
            category=UserWarning,
            message=".*Plotly can only import path collections linked to 'data' coordinates.*")
        chart = plotly.tools.mpl_to_plotly(chart)

    return chart


def _image_content_to_html(content: bytes) -> str:
    str_equivalent_image = base64.b64encode(content).decode()
    return "<img src='data:image/png;base64," + str_equivalent_image + "'/>"


def _get_numpy_as_image(array):
    data_range_warnings = []
    array_min = array.min()
    array_max = array.max()
    if array_min < 0:
        data_range_warnings.append(f"the smallest value in the array is {array_min}")
    if array_max > 1:
        data_range_warnings.append(f"the largest value in the array is {array_max}")
    if data_range_warnings:
        data_range_warning_message = (" and ".join(data_range_warnings) + ". ").capitalize()
        click.echo(
            f"{data_range_warning_message}"
            f"To be interpreted as colors correctly values in the array need to be in the [0, 1] range.",
            err=True
        )
    array *= 255
    shape = array.shape
    if len(shape) == 2:
        return _get_pil_image_data(pilimage_fromarray(array.astype(numpy_uint8)))
    if len(shape) == 3:
        if shape[2] == 1:
            array2d = numpy_array([[col[0] for col in row] for row in array])
            return _get_pil_image_data(pilimage_fromarray(array2d.astype(numpy_uint8)))
        if shape[2] in (3, 4):
            return _get_pil_image_data(pilimage_fromarray(array.astype(numpy_uint8)))
    raise ValueError("Incorrect size of numpy.ndarray. Should be 2-dimensional or"
                     "3-dimensional with 3rd dimension of size 1, 3 or 4.")


def _get_pil_image_data(image: PILImage) -> bytes:
    with io.BytesIO() as image_buffer:
        image.save(image_buffer, format='PNG')
        return image_buffer.getvalue()


def _get_figure_image_data(figure) -> bytes:
    with io.BytesIO() as image_buffer:
        figure.savefig(image_buffer, format='png', bbox_inches="tight")
        return image_buffer.getvalue()


def _is_torch_tensor(image):
    return image.__class__.__module__.startswith('torch')\
           and image.__class__.__name__ == 'Tensor'\
           and hasattr(image, "numpy")


def _is_tensorflow_tensor(image):
    return image.__class__.__module__.startswith('tensorflow.')\
           and 'Tensor' in image.__class__.__name__\
           and hasattr(image, "numpy")


def _is_matplotlib_pyplot(chart):
    return chart.__class__.__module__.startswith('matplotlib.pyplot')


def is_numpy_array(image) -> bool:
    return numpy_ndarray is not None and isinstance(image, numpy_ndarray)


def is_pil_image(image) -> bool:
    return PILImage is not None and isinstance(image, PILImage)


def is_matplotlib_figure(image):
    return image.__class__.__module__.startswith('matplotlib.') and image.__class__.__name__ == 'Figure'


def is_plotly_figure(chart):
    return chart.__class__.__module__.startswith('plotly.') and chart.__class__.__name__ == 'Figure'


def is_altair_chart(chart):
    return chart.__class__.__module__.startswith('altair.') and 'Chart' in chart.__class__.__name__


def is_bokeh_figure(chart):
    return chart.__class__.__module__.startswith('bokeh.') and chart.__class__.__name__ == 'Figure'


def is_pandas_dataframe(table):
    return isinstance(table, DataFrame)


def _export_pandas_dataframe_to_html(table):
    buffer = StringIO(table.to_html())
    buffer.seek(0)
    return buffer.getvalue()


def _export_plotly_figure(image):
    buffer = StringIO()
    image.write_html(buffer)
    buffer.seek(0)
    return buffer.getvalue()


def _export_altair_chart(chart):
    buffer = StringIO()
    chart.save(buffer, format='html')
    buffer.seek(0)
    return buffer.getvalue()


def _export_bokeh_figure(chart):
    from bokeh.resources import CDN
    from bokeh.embed import file_html

    html = file_html(chart, CDN)
    buffer = StringIO(html)
    buffer.seek(0)
    return buffer.getvalue()


def _export_pickle(obj):
    buffer = BytesIO()
    pickle.dump(obj, buffer)
    buffer.seek(0)
    return buffer.getvalue()



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging
import sys
import threading
import traceback
import uuid
from typing import TYPE_CHECKING, Callable, List
from platform import node as get_hostname

if TYPE_CHECKING:
    pass

_logger = logging.getLogger(__name__)


class UncaughtExceptionHandler:

    def __init__(self):
        self._previous_uncaught_exception_handler = None
        self._handlers = dict()
        self._lock = threading.Lock()

    def activate(self):
        with self._lock:
            this = self
            def exception_handler(exc_type, exc_val, exc_tb):
                header_lines = [
                    f"An uncaught exception occurred while run was active on worker {get_hostname()}.",
                    "Marking run as failed",
                    "Traceback:"
                ]

                traceback_lines = header_lines + traceback.format_tb(exc_tb) + str(exc_val).split("\n")
                for _, handler in self._handlers.items():
                    handler(traceback_lines)

                # pylint: disable=protected-access
                this._previous_uncaught_exception_handler(exc_type, exc_val, exc_tb)

            if self._previous_uncaught_exception_handler is None:
                self._previous_uncaught_exception_handler = sys.excepthook
                sys.excepthook = exception_handler

    def deactivate(self):
        with self._lock:
            sys.excepthook = self._previous_uncaught_exception_handler
            self._previous_uncaught_exception_handler = None

    def register(self, uid: uuid.UUID, handler: Callable[[List[str]], None]):
        with self._lock:
            self._handlers[uid] = handler

    def unregister(self, uid: uuid.UUID):
        with self._lock:
            if uid in self._handlers:
                del self._handlers[uid]


instance = UncaughtExceptionHandler()



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging
from typing import Optional

from neptune.new.internal.backends.hosted_neptune_backend import HostedNeptuneBackend
from neptune.new.internal.backends.project_name_lookup import project_name_lookup
from neptune.new.internal.credentials import Credentials
from neptune.new.internal.utils import verify_type
from neptune.new.project import Project
from neptune.new.version import version as parsed_version

__version__ = str(parsed_version)

_logger = logging.getLogger(__name__)


def get_project(name: Optional[str] = None, api_token: Optional[str] = None) -> Project:
    """Get a project with given `name`.

    Args:
        name(str, optional): Name of a project in a form of namespace/project_name. Defaults to `None`.
            If None, the value of `NEPTUNE_PROJECT` environment variable will be taken.
        api_token(str, optional): User’s API token. Defaults to `None`.
            If None, the value of `NEPTUNE_API_TOKEN` environment variable will be taken.
            .. note::
                It is strongly recommended to use `NEPTUNE_API_TOKEN` environment variable rather than placing your
                API token in plain text in your source code.

    Returns:
        ``Project``: object that can be used to interact with the project as a whole like fetching data from Runs table.

    Examples:

        >>> import neptune.new as neptune

        >>> # Fetch project 'jack/sandbox'
        ... project = neptune.get_project(name='jack/sandbox')

        >>> # Fetch all Runs metadata as Pandas DataFrame
        ... runs_table_df = project.fetch_runs_table().to_pandas()

    You may also want to check `get_project docs page`_.

    .. _get_project docs page:
       https://docs.neptune.ai/api-reference/neptune#get_project
    """
    verify_type("name", name, (str, type(None)))
    verify_type("api_token", api_token, (str, type(None)))

    backend = HostedNeptuneBackend(Credentials(api_token=api_token))
    project_obj = project_name_lookup(backend, name)

    return Project(project_obj.uuid, backend)



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from neptune import envs
from neptune.exceptions import NeptuneException, STYLES


class NeptuneApiException(NeptuneException):
    pass


class SSLError(NeptuneException):
    def __init__(self):
        super(SSLError, self).__init__('SSL certificate validation failed. Set NEPTUNE_ALLOW_SELF_SIGNED_CERTIFICATE '
                                       'environment variable to accept self-signed certificates.')


class ConnectionLost(NeptuneApiException):
    def __init__(self):
        super(ConnectionLost, self).__init__('Connection lost. Please try again.')


class ServerError(NeptuneApiException):
    def __init__(self):
        message = """
{h1}
----ServerError-----------------------------------------------------------------------
{end}
Neptune Client Library encountered an unexpected Server Error.

Please try again later or contact Neptune support.
"""
        super(ServerError, self).__init__(message.format(**STYLES))


class Unauthorized(NeptuneApiException):
    def __init__(self):
        message = """
{h1}
----Unauthorized-----------------------------------------------------------------------
{end}
You have no permission to access given resource.
    
    - Verify your API token is correct.
      See: https://docs-legacy.neptune.ai/security-and-privacy/api-tokens/how-to-find-and-set-neptune-api-token.html
      
    - Verify if you set your Project qualified name correctly
      The correct project qualified name should look like this {correct}WORKSPACE/PROJECT_NAME{end}.
      It has two parts:
          - {correct}WORKSPACE{end}: which can be your username or your organization name
          - {correct}PROJECT_NAME{end}: which is the actual project name you chose
          
    - Ask your organization administrator to grant you necessary privileges to the project
"""
        super(Unauthorized, self).__init__(message.format(**STYLES))


class Forbidden(NeptuneApiException):
    def __init__(self):
        message = """
{h1}
----Forbidden-----------------------------------------------------------------------
{end}
You have no permission to access given resource.
    
    - Verify your API token is correct.
      See: https://docs-legacy.neptune.ai/security-and-privacy/api-tokens/how-to-find-and-set-neptune-api-token.html
      
    - Verify if you set your Project qualified name correctly
      The correct project qualified name should look like this {correct}WORKSPACE/PROJECT_NAME{end}.
      It has two parts:
          - {correct}WORKSPACE{end}: which can be your username or your organization name
          - {correct}PROJECT_NAME{end}: which is the actual project name you chose
          
   - Ask your organization administrator to grant you necessary privileges to the project
"""
        super(Forbidden, self).__init__(message.format(**STYLES))


class InvalidApiKey(NeptuneApiException):
    def __init__(self):
        message = """
{h1}
----InvalidApiKey-----------------------------------------------------------------------
{end}
Your API token is invalid.
    
Learn how to get it in this docs page:
https://docs-legacy.neptune.ai/security-and-privacy/api-tokens/how-to-find-and-set-neptune-api-token.html

There are two options to add it:
    - specify it in your code 
    - set an environment variable in your operating system.

{h2}CODE{end}
Pass the token to {bold}neptune.init(){end} via {bold}api_token{end} argument:
    {python}neptune.init(project_qualified_name='WORKSPACE_NAME/PROJECT_NAME', api_token='YOUR_API_TOKEN'){end}

{h2}ENVIRONMENT VARIABLE{end} {correct}(Recommended option){end}
or export or set an environment variable depending on your operating system: 

    {correct}Linux/Unix{end}
    In your terminal run:
        {bash}export {env_api_token}=YOUR_API_TOKEN{end}
        
    {correct}Windows{end}
    In your CMD run:
        {bash}set {env_api_token}=YOUR_API_TOKEN{end}
        
and skip the {bold}api_token{end} argument of {bold}neptune.init(){end}: 
    {python}neptune.init(project_qualified_name='WORKSPACE_NAME/PROJECT_NAME'){end}
    
You may also want to check the following docs pages:
    - https://docs-legacy.neptune.ai/security-and-privacy/api-tokens/how-to-find-and-set-neptune-api-token.html
    - https://docs-legacy.neptune.ai/getting-started/quick-starts/log_first_experiment.html

{correct}Need help?{end}-> https://docs-legacy.neptune.ai/getting-started/getting-help.html
"""
        super(InvalidApiKey, self).__init__(message.format(
            env_api_token=envs.API_TOKEN_ENV_NAME,
            **STYLES
        ))


class WorkspaceNotFound(NeptuneApiException):
    def __init__(self, namespace_name):
        message = """
{h1}
----WorkspaceNotFound-------------------------------------------------------------------------
{end}
Workspace {python}{workspace}{end} not found.

Workspace is your username or a name of your team organization.
"""
        super(WorkspaceNotFound, self).__init__(message.format(
            workspace=namespace_name,
            **STYLES
        ))


class ProjectNotFound(NeptuneApiException):
    def __init__(self, project_identifier):
        message = """
{h1}
----ProjectNotFound-------------------------------------------------------------------------
{end}
Project {python}{project}{end} not found.

Verify if your project's name was not misspelled. You can find proper name after logging into Neptune UI.
"""
        super(ProjectNotFound, self).__init__(message.format(
            project=project_identifier,
            **STYLES
        ))


class PathInProjectNotFound(NeptuneApiException):
    def __init__(self, path, project_identifier):
        super(PathInProjectNotFound, self).__init__(
            "Path {} was not found in project {}.".format(path, project_identifier))


class PathInExperimentNotFound(NeptuneApiException):
    def __init__(self, path, exp_identifier):
        super().__init__(
            f"Path {path} was not found in experiment {exp_identifier}.")


class NotebookNotFound(NeptuneApiException):
    def __init__(self, notebook_id, project=None):
        if project:
            super(NotebookNotFound, self).__init__(
                "Notebook '{}' not found in project '{}'.".format(notebook_id, project))
        else:
            super(NotebookNotFound, self).__init__(
                "Notebook '{}' not found.".format(notebook_id))


class ExperimentNotFound(NeptuneApiException):
    def __init__(self, experiment_short_id, project_qualified_name):
        super(ExperimentNotFound, self).__init__("Experiment '{exp}' not found in '{project}'.".format(
            exp=experiment_short_id, project=project_qualified_name))


class ChannelNotFound(NeptuneApiException):
    def __init__(self, channel_id):
        super(ChannelNotFound, self).__init__("Channel '{id}' not found.".format(id=channel_id))


class ExperimentAlreadyFinished(NeptuneApiException):
    def __init__(self, experiment_short_id):
        super(ExperimentAlreadyFinished, self).__init__(
            "Experiment '{}' is already finished.".format(experiment_short_id))


class ExperimentLimitReached(NeptuneApiException):
    def __init__(self):
        super(ExperimentLimitReached, self).__init__('Experiment limit reached.')


class StorageLimitReached(NeptuneApiException):
    def __init__(self):
        super(StorageLimitReached, self).__init__('Storage limit reached.')


class ExperimentValidationError(NeptuneApiException):
    pass


class ChannelAlreadyExists(NeptuneApiException):
    def __init__(self, experiment_short_id, channel_name):
        super(ChannelAlreadyExists, self).__init__(
            "Channel with name '{}' already exists in experiment '{}'.".format(channel_name, experiment_short_id))


class ChannelDoesNotExist(NeptuneApiException):
    def __init__(self, experiment_short_id, channel_name):
        super(ChannelDoesNotExist, self).__init__(
            "Channel with name '{}' does not exist in experiment '{}'.".format(channel_name, experiment_short_id))


class ChannelsValuesSendBatchError(NeptuneApiException):
    @staticmethod
    def _format_error(error):
        return "{msg} (metricId: '{channelId}', x: {x})".format(
            msg=error.error,
            channelId=error.channelId,
            x=error.x)

    def __init__(self, experiment_short_id, batch_errors):
        super(ChannelsValuesSendBatchError, self).__init__(
            "Received batch errors sending channels' values to experiment {}. "
            "Cause: {} "
            "Skipping {} values.".format(
                experiment_short_id,
                self._format_error(batch_errors[0]) if batch_errors else "No errors",
                len(batch_errors)))


class ExperimentOperationErrors(NeptuneApiException):
    """Handles minor errors returned by calling `client.executeOperations`"""
    def __init__(self, errors):
        super().__init__()
        self.errors = errors

    def __str__(self):
        lines = ['Caused by:']
        for error in self.errors:
            lines.append(f'\t* {error}')
        return '\n'.join(lines)



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import unittest

from neptune.new.envs import API_TOKEN_ENV_NAME
from neptune.new.exceptions import NeptuneInvalidApiTokenException
from neptune.new.internal.credentials import Credentials

API_TOKEN = 'eyJhcGlfYWRkcmVzcyI6Imh0dHBzOi8vYXBwLnN0YWdlLm5lcHR1bmUubWwiLCJ' \
            'hcGlfa2V5IjoiOTJhNzhiOWQtZTc3Ni00ODlhLWI5YzEtNzRkYmI1ZGVkMzAyIn0='


class TestCredentials(unittest.TestCase):

    def test_should_take_default_credentials_from_env(self):
        # given
        os.environ[API_TOKEN_ENV_NAME] = API_TOKEN

        # when
        credentials = Credentials()

        # then
        self.assertEqual(API_TOKEN, credentials.api_token)

    def test_should_replace_token_from_env(self):
        # given
        os.environ[API_TOKEN_ENV_NAME] = "INVALID_TOKEN"

        # when
        credentials = Credentials(API_TOKEN)

        # then
        self.assertEqual(API_TOKEN, credentials.api_token)

    def test_raise_invalid_token(self):
        # expect
        with self.assertRaises(NeptuneInvalidApiTokenException):
            Credentials("INVALID_TOKEN")

    def test_raise_invalid_token_from_env(self):
        # given
        os.environ[API_TOKEN_ENV_NAME] = "INVALID_TOKEN"

        # expect
        with self.assertRaises(NeptuneInvalidApiTokenException):
            Credentials()
